"""
DATAQ 4108 driver level code
author: Valentyn Stadnytskyi
June 2018 - June 2018

1.0.0 - designed for usb Bulk protocol.
1.0.1 - dec is added to the communication

"""

from numpy import nan, mean, std, asarray, array, concatenate, delete, round, vstack, hstack, zeros, transpose, split
from serial import Serial

import sys
if sys.version_info[0] == 3:
    if sys.version_info[1] <= 7:
        from time import gmtime, strftime, time, sleep, clock
    else:
        from time import gmtime, strftime, time, sleep
        from time import perf_counter as clock
else:
    from time import gmtime, strftime, time, sleep, clock

import os.path
import struct
from pdb import pm

import logging

from struct import pack, unpack
from timeit import Timer
from logging import info,warn, debug, error
import traceback
__version__ = '1.0.1'

class Driver(object):

    def __init__(self):
        #tested dec 17, 2017
        self.available_ports = []
        self.dev = None

    def init(self):
        self.discover()
        self.use_port()

        info("initialization of the driver is complete")

    def get_information(self):
        if self.dev != None:
            self.dev_dict = {}
            self.epi_dict = {}
            self.dev_dict['DEV:address'] = self.dev.address
            self.dev_dict['DEV:bDeviceClass'] = self.dev.bDeviceClass
            self.dev_dict['DEV:bDescriptorType'] = self.dev.bDescriptorType
            self.dev_dict['DEV:bDeviceProtocol'] = self.dev.bDeviceProtocol
            self.dev_dict['DEV:bLength'] = self.dev.bLength
            self.dev_dict['DEV:bMaxPacketSize0'] = self.dev.bMaxPacketSize0
            self.dev_dict['DEV:bNumConfigurations'] = self.dev.bNumConfigurations
            self.dev_dict['DEV:manufacturer'] = self.dev.manufacturer
            self.dev_dict['DEV:serial_number'] = self.dev.serial_number
            self.dev_dict['DEV:speed'] = self.dev.speed
            self.dev_dict['DEV:product'] = self.dev.product

            #endpoint IN description
            self.epi_dict['EPI:bmAttributes'] = self.epi.bmAttributes
            self.epi_dict['EPI:wMaxPacketSize'] = self.epi.wMaxPacketSize
            self.epi_dict['EPI:bSynchAddress'] = self.epi.bSynchAddress
            self.epi_dict['EPI:bInterval'] = self.epi.bInterval
            self.epi_dict['EPI:bEndpointAddress'] = self.epi.bEndpointAddress
            self.epi_dict['EPI:bDescriptorType'] = self.epi.bDescriptorType
            self.epi_dict['EPI:bInterval'] = self.epi.bInterval
            self.epi_dict['EPI:bInterval'] = self.epi.bInterval

    def get_hardware_information(self):
        dic = {}
        dic[b'Device name'] = self.inquire(b'info 1 \r').split(b'info 1 ')[1][1:-1]
        dic[b'Firmware version'] = self.inquire(b'info 2 \r').split(b'info 2 ')[1][1:-1]
        dic[b'Serial Number'] = self.inquire(b'info 6 \r').split(b'info 6 ')[1][1:-1]
        dic[b'Sample Rate Divisor'] = self.inquire(b'info 9 \r').split(b'info 9 ')[1][1:-1]

        return dic

    def use_port(self, port = ''):
        import usb.core
        import usb.util


        self.dev.reset()

        # set the active configuration. With no arguments, the first
        # configuration will be the active one
        self.dev.set_configuration()

        # get an endpoint instance
        cfg = self.dev.get_active_configuration()
        intf = cfg[(0,0)]

        self.epo = usb.util.find_descriptor(
                                      intf,
                                      # match the first OUT endpoint
                                      custom_match = \
                                      lambda e: \
                                      usb.util.endpoint_direction(e.bEndpointAddress) == \
                                      usb.util.ENDPOINT_OUT)

        self.epi = usb.util.find_descriptor(
                                      intf,
                                      # match the first IN endpoint
                                      custom_match = \
                                      lambda e: \
                                      usb.util.endpoint_direction(e.bEndpointAddress) == \
                                      usb.util.ENDPOINT_IN)

        assert self.epo is not None
        assert self.epi is not None
        self.epi.wMaxPacketSize = 7200000
        self.epo.wMaxPacketSize = 7200000
        self.epi.bmAttributes = 1
        self.epi.bInterval = 100
        self.usb_buff = int(self.epi.wMaxPacketSize/100)
        #self.readall()

    def discover(self):
        """
        the function allows to discover DI-4108 device.
        returns: flag if a device(devices) are found.
        assigns: self.available_ports list entry
        [0] - COM port namer
        pvproperty(value=2.0)[1] - serial number
        """
        import usb.core
        flag = True
        self.dev = usb.core.find(idVendor=0x0683, idProduct=0x4108)
        if self.dev is None:
            raise ValueError('Device not found')
            flag = False
        return flag

    """Basic serial communication functions"""

    def read(self,N = 0):
        if N == 0:
            usb_buff = int(self.usb_buff)
        else:
            usb_buff = int(N)
        from time import sleep
        #tested dec 17, 2017
        string = b""
        timeout = 1000
        try:
            data = self.dev.read(self.epi,usb_buff,timeout)
        except:
            error(traceback.format_exc())
            data = ''

        if len(data) != 0:
            for i in data:
                string += bytes([i])
        return string

    def readall(self):
        return self.read(0)

    def write(self,command):
        try:
            self.dev.write(self.epo,command)
        except:
            error(traceback.format_exc())

    def inquire(self,command):
        self.write(command)
        res = self.read(self.usb_buff)
        return res

    def close_port(self):
        pass
    """DI-4108 Communcation functions"""

    def config_digital(self, number = 127, echo = False):
        #tested dec 17, 2017
        string = 'endo ' + str(number) + '\r'
        flag = False
        self.write(string)
        a = self.readall()
        info('%r' % a)
        if echo == True:
            if a == 'endo ' + str(number) + '\r':
                flag = True
            else:
                flag = False
        else:
            flag = None
        return flag

    def set_digital(self, number = 127, echo = False):
        #tested dec 17, 2017
        string = 'dout ' + str(number) + ' \r'
        flag = False
        self.write(string)
        if echo == True:
            a = self.readall()
            debug('%r' % a)
            if a == 'dout ' + str(number) + ' \r\x00':
                flag = True
            else:
                flag = False
        else:
            flag = None
        return flag

    def set_analog(self, echo = False, filterN = 1):
        #tested dec 17, 2017
        import traceback
        flag = False
        try:
            filterN = int(filterN)
        except:
            error(traceback.format_exc())
            filterN = 1

        self.write('filter * '+str(int(filterN))+' \r')
        # 0 - Sets all channels to read the last point; 1 - CIC filter
        self.write('slist 0 0 \r') # 256 Sets analog channel 0 to be read at +-5V, increment by 1 to go to next channel
        self.write('slist 1 1 \r') # 257 "" but channel 1
        self.write('slist 2 2 \r') # 1282 Sets to read channel 2 at +-.2V
        self.write('slist 3 3 \r') # Sets to read channel 3 at +-.2V
        self.write('slist 4 4 \r') # Sets to read channel 4 at +-.2V
        self.write('slist 5 5 \r') # Sets to read channel 5 at +-.2V
        self.write('slist 6 6 \r') # Sets to read channel 6 at +-5V
        self.write('slist 7 7 \r') # Sets to read channel 7 at +-5V
        self.write('slist 8 8 \r') # Sets to read digital input
        self.write('ps 6 \r') # set DATSQ packet size 1024 bytes page 3 di-4x08 Protocol

        #read buffer. The echoed commands need to be checked for succesful inplementation.
        #the following part was supposed to check if command has echoed back, meaning was received by the DI-4018.
        #I don't think this command is needed at all.
        a = self.readall()
        b = self.readall()
        #print('%r' % (a + b))
        if echo == True:
            if (a+b) == 'filter * 0 \r\x00slist 0 256 \rslist 1 257 \rslist 2 1282 \r\x00slist 3 1283 \r\x00slist 4 1284 \r\x00slist 5 1285 \r\x00slist 6 262 \rslist 7 263 \rslist 8 8 \r':
                flag = True
        else:
            flag = True
        return flag

    def set_acquisition_rate(self, rate = 1000, baserate = 20000, dec = 0, echo = True):
        #tested dec 17, 2017
        baserate = 20000
        if dec == 0:
            dec = int(baserate/rate)

        flag = False
        string = 'srate ' + str(int(60000000/baserate)) +' \r'
        #self.ser.write('srate 6000000\r') # Supposedly sets it to read at 50kHz, Hz=60000000/srate
        self.write(string)
        string = 'dec ' + str(dec) +' \r'
        self.write(string)
        from time import sleep
        sleep(0.2)

        #print('%r,%r' % (a,b))
        if echo == True:
            a = self.readall()
            debug(a)
            b = self.readall()
            debug(a)
            if a+b == 'srate '+ str(60000000/rate)+ '\r\x00':
                flag = True
        else:
            flag = True
        return flag

    def start_scan(self):
        info( self.write('start\r'))
        info( 'The configured measurement(s) has(have) started')

    def stop_scan(self):
        self.write('stop\r')
        info("measurement ended")

    def full_stop(self):
        self.stop_scan()
        self.close_port()

    def flush(self):
        """backcompetability with serial communication"""
        pass

    """Advance function"""
    def config_channels(self, rate = 3000, conf_digital = 127, set_digital = 127, dec = 20, baserate = 20000):
        x = self.config_digital(conf_digital, echo = True)
        a = self.set_digital(set_digital, echo = True)
        b = self.set_analog(echo = True)
        c = self.set_acquisition_rate(rate = rate, dec = dec, baserate = 20000, echo = True) #3000 gives maximum rate
        return x*a*b*c

    def get_readings(self, points_to_read = 1, to_read_analog = 8, to_read_digital = 1):
        to_read = int(to_read_analog)*2+int(to_read_digital)*2
        result = self.read(to_read*points_to_read)
        if b'stop' in result:
            flag = False
        else:
            flag = True
        try:
            data = asarray(unpack(('h'*to_read_analog+'BB')*points_to_read,result))
        except:
            error(traceback.format_exc())
            data = None
        #first N bytes (where N/2 is number of analog channels to read
        #see https://docs.python.org/2/library/struct.html for 'h' description
            #digital_data = array(unpack('B',res[-1:])[0])
        #This how we can format the integer to binary string ---->> bin(unpack('B',res[-1])[0])[2:].zfill(7).split()
        #this will unpack the very last byte as binary and make sure to
        #keep the leading zero. The very first zero 'D7' byte count 18 (see manual) is ommited.
        #will be shown as a string. We need to convert it to a numpy array for easier usage
        try:
            res = transpose(asarray(split(data,points_to_read)))
        except:
            error(traceback.format_exc())
            res = None
        return res, flag #(analog_data,digital_data)

    def blink(self):
        from time import sleep
        for i in range(8):
            self.write('led ' + str(i) + ' \r')
            sleep(1)
        self.write('led ' + str(7) + ' \r')


    """Test functions"""
    def self_test(self):
        self.inquire('led 7\r')
        self.tau = 0.001
        #dictionary with device parameters such as S\N, device name, ect
        self.dict = {}
        self.dict['Device name'] = self.inquire('info 1 \r').split('info 1 ')[1][1:-1]
        self.dict['Firmware version'] = self.inquire('info 2 \r').split('info 2 ')[1][1:-1]
        self.dict['Serial Number'] = self.inquire('info 6 \r').split('info 6 ')[1][1:-1]
        self.dict['Sample Rate Divisor'] = self.inquire('info 9 \r').split('info 9 ')[1][1:-1]
        #Useful variables
        #wait time after write (before) read. This seems to be not necessary.
        for i in self.dict.keys():
            print('%r, %r' % (i, self.dict[i]))
        print('information request complete: the DI-4108 with SN %r' %self.dict['Serial Number'])
        print('%r' % self.inquire('led 2\r'))

    def test1(self):
        self.self_test()
        self.config_channels()

        self.start_scan()
        while self._waiting()[0] <1:
            sleep(0.001)
        start_t = time()
        while time()-start_t<2:
            print("%0.5f %r %r" % (time()-start_t,self._waiting()[0],self.get_readings()))
        self.stop_scan()
        print('test 1 is Done. IN buffer has all data')
        print('data = dev.get_readings()')
        print('buffer waiting %r' % self._waiting()[0])

    def test2(self):
        self.self_test()
        self.config_channels()
        self.start_scan()
        sleep(6)
        self.stop_scan()
        sleep(1)
        while self._waiting()[0]>5:
            print('time %r and value %r'% (time(),self.get_readings()))
        print('test 2 is Done')

    def test3(self):
        self.self_test()
        self.config_channels()
        self.start_scan()
        sleep(6)
        self.stop_scan()
        print(self.waiting())
        sleep(1)
        while self.waiting()[0]>5:
            print(self.waiting()[0])
            print(self.get_readings())
        print('test 2 is Done')


    def echo_test1(self):
        self.config_channels()
        self.start_scan()
        while self.waiting()[0] <1:
            sleep(0.001)
        start_t = time()
        while time()- start_t <1:
            self.write('dout 0\r')
            self.write('dout 127 \r')
            sleep(0.06)
        self.stop_scan()
        print("%r" % self._waiting()[0])
        data = self.readall()
        print('%r' % data)

    def performance_test1(self):
        self.self_test()
        self.config_channels()

        self.start_scan()
        while self.waiting()[0] <10000:
            sleep(0.001)
        start_t = time()
        self.stop_scan()

        print('test 1 is Done. IN buffer has all data')
        print('data = dev.get_readings(10)')
        print('buffer waiting %r' % self._waiting()[0])
        print('t = Timer(lambda: dev.get_readings(N))')
        print('print t.timeit(number = M)')

    def performance_test2(self):
        self.self_test()
        self.config_channels()

        self.start_scan()
        start_t = time()
        self.lst = []
        time_st = time()
        while time()-start_t<10:
            if self._waiting()[0]> 512*16:
                data = self.get_readings(512)
                self.lst.append([time()-start_t,self._waiting()[0],(time() - time_st)*1000])
                print("%r %0.10f" % (self._waiting()[0], (time() - time_st)*1000))
            sleep(24/1000) #wait for 12.8 ms
        start_t = time()
        self.stop_scan()
        print('time between 4 kS = %0.5f' % mean(10.0/len(self.lst)))
        print('Sampling rate: %0.1f' % (512/mean(10.0/len(self.lst))))
        print('test 1 is Done. IN buffer has all data')
        print('data = dev.get_readings(10)')
        print('buffer waiting %r' % self._waiting()[0])
        print('t = Timer(lambda: dev.get_readings(N))')
        print('print t.timeit(number = M)')

driver = Driver()

if __name__ == "__main__": #for testing
    import logging
    from tempfile import gettempdir
    self = driver


    logging.basicConfig(#filename=gettempdir()+'/DI_4108_BULK_LL.log',
                        level=logging.DEBUG, format="%(asctime)s %(levelname)s: %(message)s")
    print('self.self_test()')
    print('self.test1()')
    print('self.test2()')
    print('self.echo_test1()')
    print('self.performance_test1()')
    print('self.performance_test2()')



"""
DATAQ 4108 Device Level code
author: Valentyn Stadnytskyi
Date: November 2017- April 2019

fully python3 compatible.

The main purpose of this module is to provide useful interface between DI-4108 and a server system level code that does all numbercrynching. This modules only job is to attend DI-4108 and insure that all data is collected and stored in a circular buffer.

The communication in this module is done via XLI module developed by Valentyn Stadnytskyi. This module is based on python sockets.

"""#!/usr/bin/env python3
"""
Simple IOC based on caproto library.
It has
"""
from caproto.server import pvproperty, PVGroup, ioc_arg_parser, run
import caproto
from textwrap import dedent
from pdb import pm

from numpy import random, array, zeros, ndarray, nan, isnan

import sys
if sys.version_info[0] == 3:
    if sys.version_info[1] <= 7:
        from time import gmtime, strftime, time, sleep, clock
    else:
        from time import gmtime, strftime, time, sleep
        from time import perf_counter as clock
else:
    from time import gmtime, strftime, time, sleep, clock

import numpy as np

logging_shape = (1080,768, 3)
pre_shape = (216,768, 3)
depre_shape = (216,768, 3)
period_shape = (216,768, 3)
arr_shape = (64,10)

class Server(PVGroup):
    """
    An IOC with three uncoupled read/writable PVs

    Scalar PVs
    ----------
    CPU
    MEMORY
    BATTERY

    Vectors PVs
    -----------

    """

    arr_logging = np.zeros(logging_shape).flatten() + 255
    image_logging = pvproperty(value=arr_logging, dtype = int, max_length = logging_shape[0]*logging_shape[1]*logging_shape[2])

    arr_pre = np.zeros(pre_shape).flatten() + 255
    image_pre = pvproperty(value=arr_pre, dtype = int, max_length = pre_shape[0]*pre_shape[1]*pre_shape[2])

    arr_depre = np.zeros(depre_shape).flatten() + 255
    image_depre = pvproperty(value=arr_depre, dtype = int, max_length = depre_shape[0]*depre_shape[1]*depre_shape[2])

    arr_period = np.zeros(period_shape).flatten() + 255
    image_period = pvproperty(value=arr_period, dtype = int, max_length = period_shape[0]*period_shape[1]*period_shape[2])

    t1 = pvproperty(value=1.0)
    dt = pvproperty(value=1.0, precision = 3)

    server_name = pvproperty(value='event_handler_mock')

    sample_pressure = pvproperty(value=0.0, read_only = True, dtype = float, precision = 3)
    target_pressure = pvproperty(value=0.0, read_only = True, dtype = float, precision = 3)

    pump_counter = pvproperty(value=0) #done
    valves_per_pump_current = pvproperty(value=1.0, precision = 2) #done
    valves_per_pump_total = pvproperty(value=1.0, precision = 2) #done

    ctrl_operating_mode = pvproperty(value=1)
    ctrl_shutdown_state = pvproperty(value=0,dtype = int)
    ctrl_pump_state = pvproperty(value=0,dtype = int)
    ctrl_disable_pump_state = pvproperty(value=0,dtype = int)
    ctrl_pre_state = pvproperty(value=0,dtype = int)
    ctrl_depre_state = pvproperty(value=0,dtype = int)


    table_pressure_after_pre = pvproperty(value=1.0, precision = 3)
    table_pressure_before_depre = pvproperty(value=1.0, precision = 3)
    table_time_to_switch_pre = pvproperty(value=1.0, precision = 3)
    table_time_to_switch_depre = pvproperty(value=1.0, precision = 3)
    table_rise_slope = pvproperty(value=1.0, precision = 3)
    table_fall_slope = pvproperty(value=1.0, precision = 3)
    table_pulse_width_pre = pvproperty(value=1.0, precision = 3)
    table_pulse_width_depre = pvproperty(value=1.0, precision = 3)
    table_delay = pvproperty(value=1.0, precision = 3)
    table_period = pvproperty(value=1.0, precision = 3)
    table_valve_counter_pre = pvproperty(value=1.0, precision = 3)
    table_valve_counter_depre = pvproperty(value=1.0, precision = 3)

    warning_text = pvproperty(value='this is a warning/faults field')

    @sample_pressure.startup
    async def sample_pressure(self, instance, async_lib):
        # This method will be called when the server starts up.
        self.io_pull_queue = async_lib.ThreadsafeQueue()
        self.io_push_queue = async_lib.ThreadsafeQueue()
        self.device.io_push_queue = self.io_push_queue
        self.device.io_pull_queue = self.io_pull_queue


        # Loop and grab items from the response queue one at a time
        while True:
            io_dict = await self.io_push_queue.async_get()
            # Propagate the keypress to the EPICS PV, triggering any monitors
            # along the way
            for key in list(io_dict.keys()):
                if key == 'sample_pressure':
                    await self.sample_pressure.write(io_dict[key])
                elif key == 'target_pressure':
                    await self.target_pressure.write(io_dict[key])
                elif key == 'valves_per_pump_current':
                    await self.valves_per_pump_current.write(io_dict[key])
                elif key == 'valves_per_pump_total':
                        await self.valves_per_pump_total.write(io_dict[key])
                elif key == 'pump_counter':
                    await self.pump_counter.write(io_dict[key])
                elif key == 'table_pulse_width_depre':
                    await self.table_pulse_width_depre.write(io_dict[key])
                elif key == 'table_time_to_switch_depre':
                    await self.table_time_to_switch_depre.write(io_dict[key])
                elif key == 'table_fall_slope':
                    await self.table_fall_slope.write(io_dict[key])
                elif key == 'image_depre':
                    await self.image_depre.write(io_dict[key])

                elif key == 'table_time_to_switch_pre':
                    await self.table_time_to_switch_pre.write(io_dict[key])
                elif key == 'table_rise_slope':
                    await self.table_rise_slope.write(io_dict[key])
                elif key == 'image_pre':
                    await self.image_pre.write(io_dict[key])


if __name__ == '__main__':


    ioc_options, run_options = ioc_arg_parser(
        default_prefix='event_handler_mock:',
        desc=dedent(Server.__doc__))
    ioc = Server(**ioc_options)


    ioc.device = device
    run(ioc.pvdb, **run_options)



"""
DATAQ 4108 Device Level code
author: Valentyn Stadnytskyi
Date: November 2017- November 2019

fully python3 compatible.

The main purpose of this module is to provide useful interface between DI-4108 and a server system level code that does all numbercrynching. This modules only job is to attend DI-4108 and insure that all data is collected and stored in a circular buffer.

The communication in this module is done via XLI module developed by Valentyn Stadnytskyi. This module is based on python sockets.

"""
__version__ = '0.0.0'


import traceback
import psutil, os


from numpy import nan, mean, std, nanstd, asfarray, asarray, hstack, array, concatenate, delete, round, vstack, hstack, zeros, transpose, split, unique, nonzero, take, savetxt, min, max
import sys

if sys.version_info[0] == 3:
    if sys.version_info[1] <= 7:
        from time import gmtime, strftime, time, sleep, clock
    else:
        from time import gmtime, strftime, time, sleep
        from time import perf_counter as clock
else:
    from time import gmtime, strftime, time, sleep, clock

import sys
#sys.path.append('/Users/femto-13/All-Projects-on-femto/LaserLab/Software/')

import os.path
import struct
from pdb import pm

from logging import debug,info,warning,error
# see https://vstadnytskyi.github.io/auxiliary/saved-property.html for details


if sys.version_info[0] ==3:
    keyboard_input = input
    from time import process_time as clock
else:
    keyboard_input = raw_input
    from time import clock



from struct import pack, unpack
from timeit import Timer, timeit
from circular_buffer_numpy.queue import Queue
from threading import Thread, Event, Timer, Condition
from datetime import datetime

from ubcs_auxiliary.saved_property import DataBase, SavedProperty
from ubcs_auxiliary.threading import new_thread
from ubcs_auxiliary.advsleep import precision_sleep #home-made module for accurate sleep


class DI4108_DL():
    db = DataBase(root = 'TEMP', name = 'DI4108_DL')
    pr_serial_number = SavedProperty(db,'Serial Number', '00000').init()
    pr_AI_channels = SavedProperty(db,'Number of AI',8).init()
    pr_DI_channels = SavedProperty(db,'Number of DI',1).init()
    pr_rate = SavedProperty(db,'rate(Hz)', 4000).init()
    pr_baserate = SavedProperty(db,'baserate(Hz)', 20000).init()
    pr_dec = SavedProperty(db,'dec', 5).init()
    pr_DOI_config = SavedProperty(db,'DIO config', '1111111').init()
    pr_DOI_set = SavedProperty(db,'DIO set', '1111111').init()
    pr_packet_size = SavedProperty(db,'packet size', 64).init()
    pr_packet_buffer_size = SavedProperty(db,'pr_packet_buffer_size', 1000).init()
    pr_os_buffer = SavedProperty(db,'os_buffer', 12800).init()
    pressure_sensor_offset = SavedProperty(db,'pressure_sensor_offset',
                                               [69.5595, 65.562, 68.881375, 84.2195, 86.96075, 17.248, 17.322, 0]).init() # value read at atmoshperic pressure
    pg_period = SavedProperty(db,'pg_period', 5.0).init()
    pg_pre_pulse_width = SavedProperty(db,'pg_pre_pulse_width', 0.1).init()
    pg_depre_pulse_width = SavedProperty(db,'pg_depre_pulse_width', 0.1).init()
    pg_delay_width = SavedProperty(db,'pg_delay_width', 0.5).init()
    trigger_mode = 0
    DIO_default = 126



    def __init__(self):
        self.name = 'DI4108_DL'
        self.device_info = {}
        self.device_info['empty'] = 'empty'
        self.task_name_dict = {}
        self.task_name_dict['empty'] = 'empty'
        self.pr_rate = (self.pr_baserate)/self.pr_dec
        self.pr_buffer_size = (int(self.pr_packet_size*self.pr_rate),10)
        self.queue = Queue(shape = self.pr_buffer_size, dtype = 'int16')
        self.OverflowFlag = False

        self.io_push_queue = None
        self.io_put_queue = None

    def first_time_setup(self):
        self.pr_AI_channels = 8
        self.pr_DI_channels = 1
        self.pr_packet_size = 128
        self.pr_baserate = 20000
        self.pr_dec = 5
        self.pr_rate = self.pr_baserate*1.0/self.pr_dec
        self.pr_DOI_config = '1111111' #all outputs
        self.pr_DOI_set = '1111111' #all high
        self.pressure_sensor_offset = [69.5595, 65.562, 68.881375, 84.2195, 86.96075, 17.248, 17.322, 0]

    def bind_driver(self, driver = None):
        self.driver = driver

    def _init(self):
        if self.driver is not None:
            self.driver.init()
            self.device_info = {**self.driver.get_hardware_information(),**self.device_info}
            self.stop()
            self.config_DL(baserate = self.pr_baserate,dec = self.pr_dec, DOI_config = self.pr_DOI_config, DOI_set = self.pr_DOI_set)
        else:
            warning('The driver object in the device _init() is {}'.format(driver))

    def _set_priority(self):
        import traceback
        import platform #https://stackoverflow.com/questions/110362/how-can-i-find-the-current-os-in-python
        p = psutil.Process(os.getpid()) #source: https://psutil.readthedocs.io/en/release-2.2.1/
        # psutil.ABOVE_NORMAL_PRIORITY_CLASS
        # psutil.BELOW_NORMAL_PRIORITY_CLASS
        # psutil.HIGH_PRIORITY_CLASS
        # psutil.IDLE_PRIORITY_CLASS
        # psutil.NORMAL_PRIORITY_CLASS
        # psutil.REALTIME_PRIORITY_CLASS
        try:
            if platform.system() == 'Windows':
                p.nice(psutil.REALTIME_PRIORITY_CLASS)
            elif platform.system() == 'Linux': #linux FIXIT
                p.nice(-20) # nice runs from -20 to +12, where -20 the most not nice code(highest priority)
        except:
            error(trace.format_exc())
        self.proccess = p

    def init(self, msg_in = None, client = None, driver = None):

        """
        initialize the DL program
        """
        self._init()
        self.start()
        flag = True
        buff = nan
        err = None
        return flag, buff, err

    def help(self):
        """
        returns helps string

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> device.help()
        """
        response = ''
        response += 'This is DI-4108 Data Acquisition Unit python server. \n'
        response += 'Description of the Python Code' + __doc__
        return response

    def close(self):
        """
        orderly closes the device program. just a wrapper for self.stop()

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> device.abort()
        """
        self.stop()
        flag = True
        buff = nan
        err = None
        return flag, buff, err

    def abort(self):
        """
        aborts. Currently not implemented

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> device.abort()
        """
        pass



    def snapshot(self):
        """
        returns a hardcoded snapshot from the device instance. Currently does nothing.

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> device.snapshot()
        """
        from numpy import nan
        err = ''
        flag = True
        message = ''
        response = {}
        response[b'flag'] = flag
        response[b'message'] = message
        response[b'error'] = err
        return response

    def start(self):
        """
        Creates a new thread and submits self.run for execution

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> device.start()
        """
        from ubcs_auxiliary.threading import new_thread
        new_thread(self.run)

    def stop(self):
        """
        Orderly stop of the device code.
        - stops data acquisiion
        - erases all data from USB buffers
        - set digital IO to all high
        - closes usb port connection

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> device.stop()
        """
        self.running = False
        self.driver.stop_scan()
        self.driver.flush()
        self.set_DIO('1111111') # turns all valves off
        self.driver.close_port()


    def kill(self):
        """
        Orderly self-destruction. kills all threads and deletes the instance.

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> device.kill()
        """
        self.stop()
        del self

    def config_DL(self,baserate = None,dec = None, DOI_config = None, DOI_set = None):
        """
        configures analog and digital channels, and rate via baserate and dec.

        Parameters
        ----------
        value :: string

        Returns
        -------

        Examples
        --------
        >>> device.config_DIO('00000000')
        """
        debug('DL: config DL start')
        self.driver.config_channels(baserate = baserate,
                                 dec = dec,
                                 conf_digital = int(DOI_config,2),
                                 set_digital = int(DOI_set,2))
        debug('DL: config DL end')

    def config_DIO(self, value = '00000000'):
        """
        configure direction of digital IOs. It takes a string with 8 characters, '0' or '1' where
        '0' -
        '1' -

        Parameters
        ----------
        value :: string

        Returns
        -------

        Examples
        --------
        >>> device.config_DIO('00000000')
        """
        if self.running:
            driver.config_digital(int(string,2))
        else:
            driver.config_digital(int(string,2), echo = True)

    def set_DIO(self, value = '00000000'):
        """
        sets digital input/output state

        Parameters
        ----------
        value :: string

        Returns
        -------

        Examples
        --------
        >>> device.DIO = 127
        """
        if isinstance(value,str):
            value = int(value,2)
        self.driver.set_digital(value)
        self.io_push({'dio':value})
        print('set DIO inside device daq library')

    def set_outside_DIO(self, value = '00000000'):
        """
        sets digital input/output state

        Parameters
        ----------
        value :: string

        Returns
        -------

        Examples
        --------
        >>> device.DIO = 127
        """
        if isinstance(value,str):
            value = int(value,2)
        self.driver.set_digital(value)
        print('set DIO inside device daq library')


    def get_DIO(self):
        return int(self.queue.peek_last_N(3)[-1,9])
    DIO = property(get_DIO,set_DIO)
    ### Advance function


    def run(self):
        """
        This is a main fuction that will execute run_once function on the while running loop. Usually it gets submitted to a separate thread.

        This function collects data and puts it in the Ring Buffer.
        It is run in a separate thread(See main priogram)

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> device.run()
        """
        from time import strftime, localtime, time
        info('Measurement thread has started')
        self.driver.readall()
        self.running = True
        self.driver.start_scan()
        while self.running:
            self.run_once()
        info('data acquisition thread has stopped')

    def run_once(self):
        """
        the repeated block executed in while running loop in the main run() thread

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> device.run_once()
        """
        from numpy import array
        data, flag = (self.driver.get_readings(points_to_read = self.pr_packet_size))
        self.data = data
        debug(f'data from one reading cycle of the driver {data.shape}')
        debug(f'the length of pressure sensor is  offset{len(self.pressure_sensor_offset)}')
        pressure_sensor_offset = array(self.pressure_sensor_offset)
        if flag:
            for i in range(8):
                try:
                   data[:,i] = data[:,i] - pressure_sensor_offset[i]
                except:
                    error(traceback.format_exc())
            self.queue.enqueue(data)

        else:
            info('buffer overflow')
            self.running = False


    def read_queue(self, msg_in = {b'number':0}, client = None):
        """
        returns N points from the queue
        """
        response = {}
        flag = True
        err = ''
        try:
            N = msg_in[b'number']
        except:
            N = 0
            flag = False
            err+= traceback.format_exc()
        data = self.queue.dequeue(N = N)
        response[b'message'] = data
        response[b'flag'] = flag
        response[b'err'] = err
        return response

    #####Auxiliary functions
    def parse_binary(self,value = 0):
        """
        takes an integer and converts it to 8 bit representation as an array.
        If float number is passed, it will be converted to int.
        """
        from numpy import arange,ndarray,nan
        value = int(value)
        binary = format(value, '#010b')
        arr = arange(7)
        for i in range(7):
            arr[i] = binary[9-i]
        return arr

    def unparse_binary(self,arr = array([1, 1, 1, 1, 1, 1, 1])):
        """
        takes an integer and converts it to 8 bit representation as an array.
        If float number is passed, it will be converted to int.
        """
        from numpy import arange,ndarray,nan
        integer = 0
        for i in range(len(arr)):
            integer += arr[i]*2**(i)
        return integer

    def set_pressure_sensor_offset(self,N = 3):
        """
        """
        from time import sleep
        from numpy import mean, std
        sleep(N)
        offsets = mean(self.object.queue.buffer[:,self.object.queue.rear-4000*N:self.object.queue.rear],\
                       axis = 1)
        stds = std(self.object.queue.buffer[:,self.object.queue.rear-4000*N:self.object.queue.rear],\
                       axis = 1)
        print('Pressure sensors offsets are %r and errors are %r ' %(offsets,stds))
        self.pressure_sensor_offset = offsets[:8]


    ### Input-Output controller section
    def io_push(self,io_dict = None):
        """
        a wrapper that takes care of write command to the io module

        Parameters
        ----------
        io_dict :: dictionary
            a string name of the variable

        Returns
        -------

        Examples
        --------
        >>> self.io_push()
        """
        if self.io_push_queue is not None:
            self.io_push_queue.put(io_dict)

    def io_pull(self, io_dict):
        """
        a wrapper that takes care of 'read' command to the io module

        Parameters
        ----------
        name :: string
            a string name of the variable
        value :: object
            the new value of the variable to be read from the io module

        Returns
        -------

        Examples
        --------
        >>> self.io_pull()
        """
        if self.io_push_queue is not None:
            for key, value in io_dict.items:
                print(f'received update to {key} to change to {value}')


    # tests in progress

    # Old function with unknown purpose
    def submit_DIO(self,digital):
        """
        set DIO command
        """
        self.set_DIO(digital)
        flag = True
        buff = nan
        err = None
        return flag, buff, err

    def get_all_buffer(self):
        flag = True
        buff = nan
        err = None
        buff = self.ring_buffer.buffer
        return flag, buff,err

    def grab_buffer(self,pointers = (nan,nan)):
        if pointers[1] == 0:
            pointers[1] = self.ring_buffer.pointer
        array = self.ring_buffer.buffer[0,:]
        flag = False
        err = ''
        try:
            buff = self.ring_buffer.buffer[:,pointers[0]:pointers[1]]
            flag = True
        except Exception:
            err += traceback.format_exc()
            error(traceback.format_exc())
        return flag,buff,err




if __name__ == "__main__": #for testing
    from tempfile import gettempdir
    import logging
    logging.basicConfig(filename=gettempdir()+'/di4108_device.log',
                        level=logging.INFO, format="%(asctime)s %(levelname)s: %(message)s")
    from icarus_nmr.device import DI4108_DL
    device = DI4108_DL()
    from icarus_nmr.mock_driver import Driver
    driver = Driver()
    device.bind_driver(driver)
    device.init()



"""
PyEpics additional functionality
"""
from epics.wx.wxlib import PVMixin, EpicsFunction
import wx
import epics.wx

class PVImage(wx.StaticBitmap, PVMixin):
    """ Static text for displaying a PV value,
        with callback for automatic updates
        By default the text colour will change on alarm states.
        This can be overriden or disabled as constructor
        parameters
        """
    def __init__(self, parent, pv=None, style=None, im_size = None,
                 minor_alarm="DARKRED", major_alarm="RED",
                 invalid_alarm="ORANGERED", **kw):
        self.im_size = im_size
        wstyle = wx.ALIGN_LEFT
        if style is not None:
            wstyle = style

        wx.StaticBitmap.__init__(self, parent, wx.ID_ANY,
                               style=wstyle, **kw)
        PVMixin.__init__(self, pv=pv)
        self._fg_colour_alarms = {
            epics.MINOR_ALARM : minor_alarm,
            epics.MAJOR_ALARM : major_alarm,
            epics.INVALID_ALARM : invalid_alarm }

    def _SetValue(self, value):
        "set widget label"
        from PIL import Image
        from time import time
        import io
        if value is not None:
            # I do all the stuff below to convert the array to the right format. It seems to be transmitted as 32bit instead of 8, even though i tried created it as 8 on the server. This needs to be investigated.

            raw_image = value
            im_mode = 'RGB'
            im_size = self.im_size
            img = Image.frombuffer(im_mode, im_size, raw_image, 'raw', im_mode, 0, 1)
            width, height = img.size
            bit_img = wx.Bitmap.FromBuffer(width, height, img.tobytes())
            t1 = time()
            self.SetBitmap(bit_img)
        else:
            print('_SetValue',value)

    @EpicsFunction
    def OnPVChange(self, value):
        "called by PV callback"
        import traceback
        from numpy import array
        self.pv.auto_monitor = True
        try:
            # I don't know how to work with value correctly.
            # the value of value is <array size=30000, type=ctrl_long>
            # the type of value is <class 'str'>
            if self.pv is not None:
                value = array(self.pv.get(), dtype = 'int8')
                self._SetValue(value)
        except:
            print(traceback.format_exc())



import unittest
from numpy.testing import assert_array_equal
import sys
if sys.version_info[0] < 3:
    from time import clock
else:
    from time import perf_counter as clock

class MockTest(unittest.TestCase):
    def test_init_device(self):
        from ..driver_mock import Driver
        from ..device_daq import DI4108_DL
        from time import sleep
        driver = Driver()
        device = DI4108_DL()
        device.bind_driver(driver)
        device._init()
        device.start()
        dic = {b'Device name': '4108',
                 b'Firmware version': 'VS',
                 b'Serial Number': '00000000',
                 b'Sample Rate Divisor': '60000000'}
        result = device.driver.get_hardware_information()
        self.assertEqual(dic,result)
        result = driver.get_hardware_information()
        self.assertEqual(dic,result)
        sleep(0.05)
        length1 = device.queue.length
        data,flag = device.driver.get_readings(length1)
        self.assertEqual(data.T[9].mean(),127.0)
        self.assertEqual(data.T[9].std(), 0.0)
        self.assertEqual(length1, data.T[9].shape[0])

    def test_operation1(self):
        from icarus_nmr.device_daq import DI4108_DL
        from icarus_nmr.driver_mock import Driver
        device = DI4108_DL()
        device.pr_packet_size = 64
        device.bind_driver(Driver())
        device._init()
        device.run_once()
        #the line below does not work. Need to be investigated.
        #self.assertEqual(device.DIO, 127)
        self.assertEqual(device.queue.length, 64)



"""
DATAQ 4108 Device Level code
author: Valentyn Stadnytskyi
Date: November 2017- April 2019

fully python3 compatible.

The main purpose of this module is to provide useful interface between DI-4108 and a server system level code that does all numbercrynching. This modules only job is to attend DI-4108 and insure that all data is collected and stored in a circular buffer.

The communication in this module is done via XLI module developed by Valentyn Stadnytskyi. This module is based on python sockets.

"""#!/usr/bin/env python3
"""
Simple IOC based on caproto library.
It has
"""
from caproto.server import pvproperty, PVGroup, ioc_arg_parser, run
import caproto
from textwrap import dedent
from pdb import pm

from numpy import random, array, zeros, ndarray, nan, isnan
from time import time, sleep, ctime

arr_shape = (64,10)
class Server(PVGroup):
    """
    An IOC with three uncoupled read/writable PVs

    Scalar PVs
    ----------
    CPU
    MEMORY
    BATTERY

    Vectors PVs
    -----------

    """

    freq = pvproperty(value=4000.0, read_only = True, dtype = float, precision = 1)
    dio = pvproperty(value=0, read_only = False, dtype = int)


    queue_length = pvproperty(value=0, read_only = True, dtype = int)
    data = pvproperty(value=zeros(shape = (64*10,)), read_only = True, dtype = int, max_length = 40000000)
    peek_data = pvproperty(value=zeros(shape = (6400*10,)), read_only = True, dtype = int, max_length = 40000000)
    packet_shape = pvproperty(value=arr_shape, read_only = True, dtype = int)
    LIST = pvproperty(value=[0.0,0.0,0.0,0.0])



    @freq.startup
    async def freq(self, instance, async_lib):
        # This method will be called when the server starts up.
        self.io_pull_queue = async_lib.ThreadsafeQueue()
        self.io_push_queue = async_lib.ThreadsafeQueue()
        self.device.io_push_queue = self.io_push_queue
        self.device.io_pull_queue = self.io_pull_queue

        await self.freq.write(self.device.pr_rate)




        # Loop and grab items from the response queue one at a time
        while True:
            io_dict = await self.io_push_queue.async_get()
            # Propagate the keypress to the EPICS PV, triggering any monitors
            # along the way
            print(f'inside while True {io_dict}')
            for key in list(io_dict.keys()):
                if key == 'dio':
                    await self.dio.write(io_dict[key])

                elif key == 'a':
                    await self.a.write(io_dict[key])
                elif key == 'b':
                    await self.b.write(io_dict[key])
                elif key == 'c':
                    await self.c.write(io_dict[key])
                elif key == 'd':
                    await self.d.write(io_dict[key])


    @dio.putter
    async def dio(self, instance, value):
        print('Received update for the {}, sending new value {}'.format('dio',value))
        self.device.set_outside_DIO(value = value)
        return value
    @dio.getter
    async def dio(self, instance):
        print(f"getter: {'dio'}, new value ")
        value = self.device.get_DIO()
        return value

    @data.getter
    async def data(self, instance):
        value = self.device.queue.dequeue(self.device.pr_packet_size).flatten()
        if value.shape[0] > 1280:
            print(f'{value.shape}')
            print(f"getter: {'data'}, queue length {self.device.queue.length}")
            print(f"getter: {'data'}, queue rear {self.device.queue.rear}")
        return value

    @peek_data.getter
    async def peek_data(self, instance):
        value = self.device.queue.peek_first_N(N=6400).flatten()
        return value

    @queue_length.startup
    async def queue_length(self, instance, async_lib):
        await self.queue_length.write(self.device.queue.length)
    @queue_length.getter
    async def queue_length(self, instance):
        print(f"getter: {'data'}, queue length {self.device.queue.length}")
        value = self.device.queue.length
        return value

    # @ai_offset.startup
    # async def ai_offset(self, instance, async_lib):
    #     await self.ai_offset.write(self.device.pressure_sensor_offset)
    # @ai_offset.putter
    # async def ai_offset(self, instance, value):
    #     print(f"Received update for the {'dio'}, sending new value {value}")
    #     self.device.pressure_sensor_offset = value
    #     return value
    # @ai_offset.getter
    # async def ai_offset(self, instance):
    #     print(f"getter: {'ai_offset'}, new value ")
    #     value = self.device.pressure_sensor_offset
    #     return value


if __name__ == '__main__':
    from icarus_nmr.device_daq import DI4108_DL
    device = DI4108_DL()
    device.pr_buffer_size =  (6400,10)
    from icarus_nmr.mock_driver import Driver
    driver = Driver()
    device.bind_driver(driver)
    device.init()

    ioc_options, run_options = ioc_arg_parser(
        default_prefix='device_dl:',
        desc=dedent(Server.__doc__))
    ioc = Server(**ioc_options)
    device.pr_pracket_size = 128

    ioc.device = device
    run(ioc.pvdb, **run_options)



"""

Digital Input and Output handler server

"""
from caproto.server import pvproperty, PVGroup, ioc_arg_parser, run
import caproto
from textwrap import dedent
from pdb import pm

from numpy import random, array, zeros, ndarray, nan, isnan
from time import time, sleep, ctime

arr_shape = (64,10)
class Server(PVGroup):
    """
    An IOC with three uncoupled read/writable PVs

    Scalar PVs
    ----------
    CPU
    MEMORY
    BATTERY

    Vectors PVs
    -----------

    """

    dio = pvproperty(value=127, dtype = int)
    bit0 = pvproperty(value=0, dtype = int)
    bit0_enable = pvproperty(value=0, dtype = int)
    bit1 = pvproperty(value=0, dtype = int)
    bit1_enable = pvproperty(value=0, dtype = int)
    bit2 = pvproperty(value=0, dtype = int)
    bit2_enable = pvproperty(value=0, dtype = int)
    bit3 = pvproperty(value=0, dtype = int)
    bit3_enable = pvproperty(value=0, dtype = int)
    bit4 = pvproperty(value=0, dtype = int)
    bit4_enable = pvproperty(value=0, dtype = int)
    bit5 = pvproperty(value=0, dtype = int)
    bit5_enable = pvproperty(value=0, dtype = int)
    bit6 = pvproperty(value=0, dtype = int)
    bit6_enable = pvproperty(value=0, dtype = int)

    shutdown_state = pvproperty(value=0, dtype = int)

    operating_mode = pvproperty(value=0,dtype=int)

    @dio.startup
    async def dio(self, instance, async_lib):
        # This method will be called when the server starts up.
        self.io_pull_queue = async_lib.ThreadsafeQueue()
        self.io_push_queue = async_lib.ThreadsafeQueue()
        try:
            self.device.io_push_queue = self.io_push_queue
            self.device.io_pull_queue = self.io_pull_queue
        except:
            pass

        # Loop and grab items from the response queue one at a time
        while True:
            io_dict = await self.io_push_queue.async_get()
            # Propagate the keypress to the EPICS PV, triggering any monitors
            # along the way
            print(f'inside server while True loop: {io_dict}')
            for key in list(io_dict.keys()):
                if key == 'dio':
                    new_val = io_dict[key]
                    cur_val = self.dio.value
                    print(new_val,cur_val)
                    if new_val != cur_val:
                        await self.dio.write(val)
                        _bit0 = int(((val)&int(0b1))/(int(0b1)))
                        _bit1 = int(((val)&int(0b10))/(int(0b10)))
                        _bit2 = int(((val)&int(0b100))/(int(0b100)))
                        await self.bit0.write(_bit0)
                        await self.bit1.write(_bit1)
                        await self.bit2.write(_bit2)
                elif key == 'bit0':
                    await self.bit0.write(io_dict[key])
                elif key == 'bit0_enable':
                    await self.bit0_enable.write(io_dict[key])
                elif key == 'bit1':
                    await self.bit1.write(io_dict[key])
                elif key == 'bit1_enable':
                    await self.bit1_enable.write(io_dict[key])
                elif key == 'bit2':
                    await self.bit2.write(io_dict[key])
                elif key == 'bit2_enable':
                    await self.bit2_enable.write(io_dict[key])
                elif key == 'bit3':
                    await self.bit3.write(io_dict[key])
                elif key == 'bit3_enable':
                    await self.bit3_enable.write(io_dict[key])
                elif key == 'bit4':
                    await self.bit4.write(io_dict[key])
                elif key == 'bit4_enable':
                    await self.bit4_enable.write(io_dict[key])
                elif key == 'bit5':
                    await self.bit5.write(io_dict[key])
                elif key == 'bit5_enable':
                    await self.bit5_enable.write(io_dict[key])
                elif key == 'bit6':
                    await self.bit6.write(io_dict[key])
                elif key == 'bit6_enable':
                    await self.bit6_enable.write(io_dict[key])


    @dio.putter
    async def dio(self, instance, value):
        print('{}, new value {}'.format('dio putter',value))
        new_val = value
        cur_val = self.dio.value
        print(new_val,cur_val)
        if new_val != cur_val:
            _bit0 = int(((value)&int(0b1))/(int(0b1)))
            _bit1 = int(((value)&int(0b10))/(int(0b10)))
            _bit2 = int(((value)&int(0b100))/(int(0b100)))
            await self.bit0.write(_bit0)
            await self.bit1.write(_bit1)
            await self.bit2.write(_bit2)
        return value
    @operating_mode.putter
    async def operating_mode(self, instance, value):
        print('Received update for the {}, sending new value {}'.format('operating_mode',value))
        #self.device.set_DIO(value = value)
        return value
    @shutdown_state.putter
    async def shutdown_state(self, instance, value):
        print('Received update for the {}, sending new value {}'.format('shutdown_state',value))
        #self.device.set_DIO(value = value)
        return value
    @bit0.putter
    async def bit0(self, instance, value):
        print('Received update for the {}, sending new value {}'.format('bio0',value))
        cur_d = self.dio.value
        new_d = new_digital(cur_d,0,value)
        if new_d!=cur_d:
            self.device.set_dio(value = new_d)
        return value
        return value
    @bit0_enable.putter
    async def bit0_enable(self, instance, value):
        print('Received update for the {}, sending new value {}'.format('bio0_enable',value))
        #self.device.set_DIO(value = value)
        return value
    @bit1.putter
    async def bit1(self, instance, value):
        print('Received update for the {}, sending new value {}'.format('bio1',value))
        cur_d = self.dio.value
        new_d = new_digital(cur_d,1,value)
        if new_d!=cur_d:
            self.device.set_dio(value = new_d)
        return value
    @bit1_enable.putter
    async def bit1_enable(self, instance, value):
        print('Received update for the {}, sending new value {}'.format('bio1_enable',value))
        #self.device.set_DIO(value = value)
        return value
    @bit2.putter
    async def bit2(self, instance, value):
        print('Received update for the {}, sending new value {}'.format('bio2',value))
        cur_d = self.dio.value
        new_d = new_digital(cur_d,2,value)
        if new_d!=cur_d:
            self.device.set_dio(value = new_d)
        return value
    @bit2_enable.putter
    async def bit2_enable(self, instance, value):
        print('Received update for the {}, sending new value {}'.format('bio2_enable',value))
        #self.device.set_DIO(value = value)
        return value

    # @dio.getter
    # async def dio(self, instance):
    #     print(f"getter: {'dio'}, new value ")
    #     value = self.device.get_DIO()
    #     return value
    #
    # @data.getter
    # async def data(self, instance):
    #     value = self.device.queue.dequeue(self.device.pr_packet_size).flatten()
    #     if value.shape[0] > 1280:
    #         print(f'{value.shape}')
    #         print(f"getter: {'data'}, queue length {self.device.queue.length}")
    #         print(f"getter: {'data'}, queue rear {self.device.queue.rear}")
    #     return value
    #
    # @peek_data.getter
    # async def peek_data(self, instance):
    #     value = self.device.queue.peek_first_N(N=6400).flatten()
    #     return value
    #
    # @queue_length.startup
    # async def queue_length(self, instance, async_lib):
    #     await self.queue_length.write(self.device.queue.length)
    # @queue_length.getter
    # async def queue_length(self, instance):
    #     print(f"getter: {'data'}, queue length {self.device.queue.length}")
    #     value = self.device.queue.length
    #     return value

    # @ai_offset.startup
    # async def ai_offset(self, instance, async_lib):
    #     await self.ai_offset.write(self.device.pressure_sensor_offset)
    # @ai_offset.putter
    # async def ai_offset(self, instance, value):
    #     print(f"Received update for the {'dio'}, sending new value {value}")
    #     self.device.pressure_sensor_offset = value
    #     return value
    # @ai_offset.getter
    # async def ai_offset(self, instance):
    #     print(f"getter: {'ai_offset'}, new value ")
    #     value = self.device.pressure_sensor_offset
    #     return value

def int_to_binary(value):
    return bin(value+128)[-7:][::-1]

def new_digital(old_digital, bit, new_bit_value):
    """

    """
    from numpy import sign
    old_bit_value = int(int_to_binary(old_digital)[bit])
    bit_diff = new_bit_value - old_bit_value
    new_digital = old_digital + sign(bit_diff)*(2**bit)
    return new_digital

if __name__ == '__main__':

    ioc_options, run_options = ioc_arg_parser(
        default_prefix='digital_handler_mock:',
        desc=dedent(Server.__doc__))
    ioc = Server(**ioc_options)


    ioc.device = device
    run(ioc.pvdb, **run_options)



#!/bin/env python
"""
"""

class Client():
    def __init__(self,device_ca_server_prefix = 'device_mock:'):
        """
        ToDo: rewrite the function to include the list of PVs.
        - restructure subscription to PVs to be saved as a dictionary. this will allow potential future expansions.
        """
        from caproto.threading.client import Context
        self.ctx = Context()
        self.ca_name = device_ca_server_prefix
        self.dio, = self.ctx.get_pvs(f'{self.ca_name}dio',)

    def get_dio(self):
        """
        a wrapper to get digital state from the device handler process

        Parameters
        ----------

        Returns
        -------
        value :: integer

        Examples
        --------
        >>> value = client.get_dio()
        """
        return self.dio.read().data[0]

    def set_dio(self, value):
        """
        a wrapper to get digital state from the device handler process

        Parameters
        ----------
        value :: integer

        Returns
        -------

        Examples
        --------
        >>> client.set_dio(127)
        """
        result = self.dio.write(value)



#!/bin/env python
"""
The Event Detector LL for the High Pressure Apparatus
author: Valentyn Stadnytskyi
dates: June 09 2018 - November 16 2018
"""

class Handler():

    bit_HPpump = 0b1
    bit_valve1 = 0b10
    bit_valve2 = 0b100
    bit_valve3 = 0b1000
    bit_log = 0b10000
    flag_play_sequence = False
    sequence_string = 'i4L,i1L,w_150_ms,i1H,w_300_ms,i2L,w_30_ms,i2H,w_100_ms,i3L,w_500_ms,i3H,w_2_s'
    pr_pulse_generator_counter = 0


    def __init__(self, client):
        """
        """
        self.name = 'PulseGenerator'
        self.client = client
        self.running = False
        self.console_running = False

        self.io_push_queue = None
        self.io_put_queue = None

    def init(self):
        """
        initialization functions. sets instance to default values.

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> pulse_generator.init()
        """
        from numpy import nan
        self.trigger_mode = 2 # 0 - manual, 1 - pulsed, 2 - console. Always start with concole

        self.running = False

        #self.curr_DIO = int('1111111',2)


    def set_dio(self, value):
        """
        a wrapper to set digital state in the device handler process via channel access

        Parameters
        ----------
        value :: integer

        Returns
        -------

        Examples
        --------
        >>> value = client.get_dio()
        """

        self.client.set_dio(value)


    def get_dio(self):
        """
        a wrapper to get digital state in the device handler process via channel access

        Parameters
        ----------

        Returns
        -------
        value :: integer

        Examples
        --------
        >>> value = self.client.get_dio()
        """

        value = self.client.get_dio()

        return value


    def pulse(self, current_dio, valve = 1, duration = 1, precision = False, echo = False):
        """
        generates a negative pulse (high-low-high) for 'duration' seconds for 'valve' (bit)

        Parameters
        ----------
        current_dio :: int
        valve :: int
        duration :: float
        precision :: boolean
        echo :: boolean

        Returns
        -------

        Examples
        --------
        >>> self.pulse(valve = 1, duration = 1)
        """
        from time import sleep
        if valve == 1:
            valve_bit = int(self.bit_valve1) # Depressure
        elif valve == 2:
            valve_bit = int(self.bit_valve2) # Pressure
        elif valve == 3:
            valve_bit = int(self.bit_valve3) # Bit 3
        elif valve == 0:
            valve_bit = int(self.bit_HPpump) # High pressure valve
        elif valve == -1:
            valve_bit = int('0000000',2) # Wait
        else:
            valve_bit = int('0000000',2) # Wait if valve selection is not recognized
        curr_DIO = current_dio
        new_DIO = curr_DIO-valve_bit
        if valve_bit != 0:
            self.set_dio(new_DIO) #down
        if precision:
            precision_sleep.psleep(duration)
        else:
            sleep(duration)
        if valve_bit != 0:
            self.set_dio(curr_DIO) # up
        print('DIO %r -> %r -> %r' % (curr_DIO,new_DIO,curr_DIO))


    def set_bit(self, current_dio, bit = 1, value = 1):
        """
        sets bit 'bit' to value 'value' (0 or 1)

        Parameters
        ----------
        bit :: int
        valve :: int

        Returns
        -------

        Examples
        --------
        >>> self.set_bit(bit = 1, value = 1)
        """
        from time import sleep
        if bit == 1:
            _bit = int(self.bit_valve1) # Depressure
        elif bit == 2:
            _bit = int(self.bit_valve2) # Pressure
        elif bit == 3:
            _bit = int(self.bit_valve3) # Bit 3
        elif bit == 0:
            _bit = int(self.bit_HPpump) # High pressure valve
        elif bit == -1:
            _bit = int('0000000',2) # Wait
        else:
            _bit = int('0000000',2) # Wait if valve selection is not recognized
        curr_DIO = current_dio
        new_DIO = curr_DIO-_bit

        self.set_dio(new_DIO) #down
        print('setting bit %r -> %r -> %r' % (curr_DIO,new_DIO))

    def pulse_sequence(self,pvname = '',value = '', char_val = ''):
        """
        the pulse sequence:
            Depressure for tw1 miliseconds
            Wait for td miliseconds
            Pressure for tw2 miliseconds
        input: None; Output: None

        Because we acquire data at 1kHz with packet size of 128, which make is
        128 ms, we have to get current state at the very beggining of the
        sequence
        """
        self.pulse(valve = 1, duration = self.tw1)
        self.pulse(valve = -1, duration = (self.td-self.tw1)) #-1 stands for waiting
        self.pulse(valve = 2, duration = self.tw2)
        self.pr_pulse_generator_counter += 1

    def set_pump_state(self,value = None):
        """
        pump is open if value = 0 and closed if value = 1
        """
        from XLI.server_LL import server
        self.curr_DIO = int(self.get_DIO())
        try:
            if value*(self.bit_HPpump) != self.curr_DIO & self.bit_HPpump:
                if value == 1:
                    self.set_DIO(self.curr_DIO + self.bit_HPpump)
                    self.curr_DIO = self.curr_DIO + self.bit_HPpump
                elif value == 0:
                    self.set_DIO(self.curr_DIO - self.bit_HPpump)
                    self.curr_DIO = self.curr_DIO - self.bit_HPpump
                    self.set_safe_state(value = 0) #Problem is here!
        except:
            error(traceback.format_exc())

    def get_pump_state(self):
        return int((self.get_DIO()& self.bit_HPpump) / self.bit_HPpump)

    def set_pressurize_state(self,value = ''):

        bit = self.bit_valve2
        self.curr_DIO = int(self.get_DIO())
        #print('current_digital',current_digital)
        try:
            if value*(bit) != self.curr_DIO & bit:
                if value == 1:
                    self.set_DIO(self.curr_DIO + bit)
                    self.curr_DIO = self.curr_DIO + bit
                elif value == 0:
                    self.set_DIO(self.curr_DIO - bit)
                    self.curr_DIO = self.curr_DIO - bit
        except:
            error(traceback.format_exc())
        #if len(pvname) == 0:
            #casput(socket_server.CAS_prefix + 'pressurize_state',value)
    def get_pressurize_state(self):
        from XLI.auxiliary import parse_binary,unparse_binary
        arr = parse_binary(self.get_DIO())
        return arr[2]

    def set_depressurize_state(self,value = None):
        from XLI.auxiliary import parse_binary,unparse_binary
        bit = self.bit_valve1
        curr_DIO = int(self.get_DIO())
        arr = parse_binary(curr_DIO)
        #print('current_digital',current_digital)
        if value == 0 or value == 1:
            arr[1] = value
        number = unparse_binary(arr)
        self.set_DIO(number)

    def get_depressurize_state(self):
        from XLI.auxiliary import parse_binary,unparse_binary
        arr = parse_binary(self.get_DIO())
        return arr[1]

    def set_log_bit_state(self,value = None):
        """
        sets the bit 4 to the passed value. Bit 4 is logging bit.
        """
        from XLI.auxiliary import parse_binary,unparse_binary
        curr_DIO = int(self.get_DIO())
        arr = parse_binary(curr_DIO)
        #print('current_digital',current_digital)
        if value == 0 or value == 1:
            arr[4] = value
        number = unparse_binary(arr)
        self.set_DIO(number)
        server.push_subscribed_updates(controls = [b'log_bit_state'])

    def get_log_bit_state(self):
        from XLI.auxiliary import parse_binary,unparse_binary
        arr = parse_binary(self.get_DIO())
        return arr[4]

    def pressurize_pulse(self,pvname = '',value = '', char_val = ''):
        #self.curr_DIO = self.get_DIO()
        self.pulse(valve = 2, duration = self.tw2/1000.)



    def depressurize_pulse(self,pvname = '',value = '', char_val = ''):
        #self.curr_DIO = self.get_DIO()
        self.pulse(valve = 1, duration = self.tw1/1000.)

    def start(self):
        """
        for back competibility with older SL codes does nothing
        """
        from ubcs_auxiliary.threading import new_thread
        if not self.running:
            new_thread(self.run)

    def stop(self):
        self.set_DIO(self.default_state)
        self.running = False

    def close(self):
        self.stop()

    def run_once(self):
        """
        """
        from icarus_nmr.event_detector import event_detector
        #print(pulse_generator.console_counter)
        if self.trigger_mode == 0:
            if not manual_flag:
                info('pulse Generator thread: set pump default')
                #self.set_pump_to_default()
                manual_flag = True
            self.console_counter = 0
            sleep(self.update_period)
            self.shutdown_flag = False
        elif self.trigger_mode == 1:
            if manual_flag:
                self.DIO = self.DIO_default
                manual_flag = False
            #pulse_valve3_duration = 0.02
            #self.pulse(valve = 3, duration = pulse_valve3_duration)#comment this
            self.length_of_pulse_sequence = self.td + self.tw2
            self.pulse_sequence() #uncomment this

            time_to_sleep = self.period -self.length_of_pulse_sequence
            Nsteps = time_to_sleep/self.update_period
            tstart = time()
            time_left = time_to_sleep - time() + tstart
            while time_left>0:
                if self.trigger_mode == 0:
                    time_left = -1
                    break
                if time_left> self.update_period:
                    sleep(self.update_period)
                else:
                    precision_sleep.psleep(time_left)
                time_left = time_to_sleep - time() + tstart

            self.console_counter = 0
        elif self.trigger_mode == 2:
            if manual_flag:
                self.set_pump_state(value = 0)
                manual_flag = False
            else:
                pass
            self.length_of_pulse_sequence = self.td + self.tw2
            time_to_sleep = self.period
            Nsteps = time_to_sleep/self.update_period
            tstart = time()
            time_left = time_to_sleep - time() + tstart
            while time_left>0:
                if (self.shutdown_flag or self.trigger_mode == 0):
                    time_left = -1
                    break
                if time_left> self.update_period and time_left>0:
                    sleep(self.update_period)
                else:
                    precision_sleep.psleep(time_left)
                time_left = time_to_sleep - time() + tstart
        elif self.trigger_mode == 3:
            try:
                lst = self.decode_sequence(self.sequence_string)
                self.play_sequence(lst = lst)
            except:
                sleep(self.update_period)
                error(traceback.format_exc())

    def run(self):
        """
        the while True loop with pulse generator.
        it has to run all the time for high pressure pump pulse to
        be generated properly.
        important parameters: trigger: 0 manual, 1 -internal; 2 - external
        """
        from time import clock, time


        #preparation of the while true loop
        self.set_DIO(self.object.ctrls.DIO_default)
        sleep(0.3)
        self.curr_DIO  = self.get_DIO()
        self.running = True
        self.paused = False
        self.meas = {}
        self.meas['period'] = self.period
        manual_flag = True
        self.console_counter = 0
        self.update_period = 0.5

        #While true loop for the device
        while self.running:
            self.run_once()





        self.running = False
        self.paused = True


    def get_deprepulse_width(self):
        return self.tw1
    def set_deprepulse_width(self,value = ''):
        value = float(value)
        self.tw1 = value
        #if len(pvname) == 0:
            #casput(socket_server.CAS_prefix+'PG_deprepulse_width',value)

    def get_delay_width(self):
        return self.td
    def set_delay_width(self,value = ''):
        value = float(value)
        self.td = value


    def get_prepulse_width(self):
        return self.tw2
    def set_prepulse_width(self,value = ''):

        value = float(value)

        self.tw2 = value
        #if len(pvname) == 0:
            #casput(socket_server.CAS_prefix+'PG_prepulse_width',value)

    def get_period(self):
        return self.object.ctrls.pg_period
    def set_period(self,value = 0):
        self.object.ctrls.pg_period = value
    period = property(get_period,set_period)

    def get_tw1(self):
        return self.object.ctrls.pg_depre_pulse_width
    def set_tw1(self,value = 0):
        self.object.ctrls.pg_depre_pulse_width = value
    tw1 = property(get_tw1,set_tw1)

    def get_tw2(self):
        return self.object.ctrls.pg_pre_pulse_width
    def set_tw2(self,value = 0):
        self.object.ctrls.pg_pre_pulse_width = value
    tw2 = property(get_tw2,set_tw2)

    def get_td(self):
        return self.object.ctrls.pg_delay_width
    def set_td(self,value = 0):
        self.object.ctrls.pg_delay_width = value
    td = property(get_td,set_td)

    def get_safe_state(self):
        return self.object.ctrls.safe_state
    def set_safe_state(self,value = 0):
        self.object.ctrls.safe_state = value
    safe_state = property(get_safe_state,set_safe_state)




    def get_trigger_mode(self):
        """
        returns trigger mode from pulse_generator instance
        """
        return self.object.ctrls.trigger_mode
    def set_trigger_mode(self,value = ''):
        """
        sets trigger mode either from CA monitor or manually via value variable
        """
        value = int(value)
        if value == 0 or value == 1 or value == 2 or value == 3:
            curr_trigger_mode = self.trigger_mode
            self.object.ctrls.trigger_mode = value
    trigger_mode = property(get_trigger_mode,set_trigger_mode)








    def decode_sequence(self,string):
        string1 = string
        string1 = string1.replace(',','","')
        string1 = string1.replace('[','["')
        string1 = string1.replace('["["','[["')
        string1 = string1.replace('["["["','[[["')
        string1 = string1.replace(']','"]')
        string1 = string1.replace(']"]',']]')
        string1 = string1.replace(']"]"]',']]]')
        return eval(string1)

    def play_sequence(self,lst):
        """
        bit_HPpump = persistent_property('bit_HPpump',0b1)
        bit_valve1 = persistent_property('bit_valve1',0b10)
        bit_valve2 = persistent_property('bit_valve2',0b100)
        bit_valve3 = persistent_property('bit_valve3',0b1000)
        bit_log = persistent_property('bit_log',0b10000)
        """
        from time import time, sleep
        flag = True
        for item in lst:
            if flag:
                if (item[0] == 'i' or item[0] == 'I'):
                    if (item[2] =='L' or item[2] =='l'):
                        value = 0
                    elif (item[2] == 'H' or item[2] == 'h'):
                        value = 1
                    else:
                        value = 0
                    if item[1] == '1':
                        self.set_depressurize_state(value = value)
                    elif item[1] == '2':
                        self.set_pressurize_state(value = value)
                    elif item[1] == '3':
                        self.set_pump_state(value = value)
                    elif item[1] == '4':
                        self.set_log_bit_state(value = value)
                    elif item[1] == '5':
                        pass#self.set_log_bit_state(pvname = ' ', value = value)
                    else:
                        pass

                elif item[0] == 'w':
                    local_lst = item.split('_')
                    if len(local_lst) == 3:
                        if local_lst[2] == 'ms':
                            gain = 10**-3
                        elif local_lst[2] == 's':
                            gain = 1
                        else:
                            gain = 1
                        try:
                            flt = float(local_lst[1])
                        except:
                            error(traceback.format_exc())
                            flt = 1
                    time_to_sleep = flt*gain
                    #precision_sleep.psleep(time_to_sleep)
                    tstart = time()
                    time_left = time_to_sleep - time() + tstart
                    while time_left>0:
                        if (self.trigger_mode != 3):
                            flag = False
                            time_left = 0.01
                        precision_sleep.psleep(time_left)
                        time_left = time_to_sleep - time() + tstart
                else:
                    pass
        if flag:
            self.set_trigger_mode(value = 0)



#from icarus_SL import icarus_SL#!/bin/env python
"""
The Event Detector LL for the High Pressure Apparatus
author: Valentyn Stadnytskyi
dates: June 09 2018 - November 16 2018

1.0.0 - the improved copy from the old class from the older SL
1.1.0 - Bugs fixed and Fully tested version that works with SL server! Exciting
      - EvalEvent function was changed. The list is cleared at the end of analysis.

Events:
two numbers: first D channel number and second direction 0(down) and 1(up)
00 - D0 goes low
01 - D0 goes high
10 - D1 goes low
11 - D1 goes high
20 - D2 goes low
21 - D2 goes high
and so on

3 digital numbers starting with 100 are analog events:
100 - analog event of high pressure pump stroke

version 3.0.0 - added tube_length and medium
 - 4.0.0 - majopr upgrade to Python 3 and minor restructuring

"""
__version__ = '4.0.0' #
import sys

import traceback

import sys
if sys.version_info[0] ==3:
    from time import time,sleep,clock
else:
    from time import time,sleep,perf_counter

from logging import debug,info,warn,error
from numpy import nan, std, inf, nanmean, nanstd, nonzero, zeros, nanargmax, nanargmin, nanmin, nanmax, asarray
import platform
import pickle
import traceback
import scipy.stats
from scipy.interpolate import UnivariateSpline

from pdb import pm

class Client():
    def __init__(self,device_ca_server_prefix = 'device_mock:', dio_ca_server_prefix = 'digital_handler_mock:'):
        """
        ToDo: rewrite the function to include the list of PVs.
        - restructure subscription to PVs to be saved as a dictionary. this will allow potential future expansions.
        """
        from caproto.threading.client import Context
        self.ctx = Context()
        self.ca_name = device_ca_server_prefix
        self.dio_ca_name = dio_ca_server_prefix
        self.dio,self.data,self.peek_data,self.queue_length,self.packet_shape,self.freq = self.ctx.get_pvs(f'{self.ca_name}dio',
            f'{self.ca_name}data',
            f'{self.ca_name}peek_data',
            f'{self.ca_name}queue_length',
            f'{self.ca_name}packet_shape',
            f'{self.ca_name}freq',)

        self.dio,= self.ctx.get_pvs(f'{self.dio_ca_name}dio',)

    def get_dio(self, value):
        return self.dio.read().data[0]
    def set_dio(self, value):
        self.dio.write(value)

if __name__ == "__main__":
    from importlib import reload
    from tempfile import gettempdir
    import logging
    import matplotlib
    matplotlib.use('WxAgg')
    client = Client()
    daq = DAQ(client)
    daq.init()
    event_detector = Event_Detector()
    logging.basicConfig(filename=gettempdir()+'/di4108_DL.log',
                        level=logging.INFO, format="%(asctime)s %(levelname)s: %(message)s")

    event_detector.init()
    from numpy import array
    self = event_detector
    temp_lst = {b'tSwitchPressure_0':1,b'tSwitchPressure_1':2,b'tSwitchPressureEst_0':3,b'gradientPressure_0':4,b'gradientPressure_1':5,b'gradientPressureEst_0':6}



"""
Icarus High Pressure Jump log files analysis software
author: Valentyn Stadnytskyi
Date: June 2019- July 2019

How it supposed to work.

The class Dataset points at a logging directory.
The directory will have 1 folder:
- buffer_files
    - time_stamp_period_type-of-trace.csv (example: 1564501923.8782883_0_depre.csv
        -> where time = 1564501923.8782883, period = 0 and type-of-data = depre (depressure event))
and 3 files
- experiment.log
- experiment.pickle - a pickled version of processed experiment.log file.
- experiment_parameters.log - a dump of all variable from all modules mostly for debugging purposes.
"""

from numpy import gradient, transpose, genfromtxt, nanmean, argwhere, where, nan, isnan, asarray
import os
import pickle
from logging import debug, info, warn, error

class Dataset():
    def __init__(self, folder = None):
        """

        :param folder: folder, defaults to None
        :type folder: str
        :param log_header: log_header, defaults to None
        :type log_header: list

        """
        ### Log folder related
        self.folder = folder #the pointer at the log folder
        ###Log file related
        self.log_header = None #the header retrieved from experiment.log
        self.log_data = None
        self.log_raw_data = None
        self.log_length = 0
        self.description = ''
        #Traces
        #a dictionary of length for each trace type.
        self.trace_length = {'pre':None, 'depre':None, 'period': None, 'pump':None, 'meanbit3': None, 'cooling': None}
        #a dictionary of lists of different trace types.
        self.trace_lists = {'pre':None, 'depre':None, 'period': None, 'pump':None, 'meanbit3': None, 'cooling': None}
        #auxiliary variables
        self.i = 0
        if folder is not None:
            self.init()

    def init(self):
        from time import time
        import os
        t1 = time()
        self.is_init_done = False
        if self.folder is not None:
            #check if the folder path ends with '/' and add if it is not.
            if self.folder[-1] != '/':
                self.folder += '/'
            #get the list of all trace types and put in appropriate dictionary entry
            for key in self.trace_lists.keys():
                self.trace_lists[key] = self.get_lst(self.folder, key)
                self.trace_length[key] = len(self.trace_lists[key])
            #get log_header
            self.log_header = self.log_read_header(self.folder)
            #check if pickle file exist and load it
            self.log_rawdata,self.log_data = self.log_read_file(self.folder)

            self.log_length = self.log_data.shape[0]
            self.is_init_done = True
        else:
            self.is_init_done  = False

        t2 = time()
        dt = t2-t1
        print('Init of Dataset from folder = {} has status {}. It took {} seconds.'.format(self.folder,self.is_init_done,dt))

    # def dump_to_pickle_file(self, data = None):
    #     """
    #     pickles data and puts it on the drive. the default name is experiment.pickle (similar to original experiment.log)
    #
    #
    #     Parameters
    #     ----------
    #     data: numpy array
    #         data to append
    #
    #     Returns
    #     -------
    #
    #     Examples
    #     --------
    #     >>> dataset = Dataset()
    #     >>> dataset.dump_to_pickle_file(data = data)
    #
    #     """
    #     from pickle import dump, HIGHEST_PROTOCOL
    #     if data is not None:
    #         with open(self.folder + 'experiment.pickle', 'wb') as file:
    #             dump(obj = data, file = file, protocol=HIGHEST_PROTOCOL)

    def log_read_header(self, folder):
        """
        looks for experiment.log file in the specified folder. Reads it and returns header. The typical folder name is /YEAR-MM-DD-hh-mm-ss, where MM - month, DD - day, hh - hours(24 hours), mm-minutes, ss-seconds

        Parameters
        ----------
        folder: string
            folder name

        Returns
        -------
        data: numpy array
            data to append

        Examples
        --------
        >>> dataset = Dataset()
        >>> folder = '/2019-05-31-13-13-52/'
        >>> header = dataset.log_read_header(folder = folder)
        >>> header

        """
        header = ""
        with open(folder + 'experiment.log', "r") as f:
            a = f.readline()
            a = f.readline()
        header = a.replace("b'","").replace("'","").replace("\n","").replace(" ","").split(',')

        return header

    def log_read_file(self, folder):
        """
        converts raw log file data to collapsed 2D numpy array where every entry corresponds to one period. The time stamp on the period will be taken from the period event, which is defined elsewhere.

        looks for experiment.log file in the specified folder. Reads it and returns data as numpy array. The typical folder name is /YEAR-MM-DD-hh-mm-ss, where MM - month, DD - day, hh - hours(24 hours), mm-minutes, ss-seconds

        Parameters
        ----------
        folder: string
            folder name

        Returns
        -------
        data: numpy array
            data to append

        Examples
        --------
        >>> folder = '/2019-05-31-13-13-52/'
        >>> raw_data, data = dataset.log_read_file(folder = folder)
        >>> raw_data.shape
        (980, 37)
        >>> data.shape
        (403, 37)

        """
        from os.path import exists
        from numpy import genfromtxt
        raw_data = genfromtxt(folder + 'experiment.log', delimiter = ',', skip_header = 2)
        if exists(folder + 'experiment.pickle'):
            print('the experiment.pickle file in the folder {} was detected and will be uploaded'.format(folder))
            # read processed data from .pickle file
            data = self.log_load_pickle_file(folder)

        else:
            info('the pickle file in the folder {} was NOT detected and the processing of the raw .log file is initiated'.format(folder))
            data =  raw_data
            self.dump_to_pickle_file(data)

        return raw_data, data

    def log_load_pickle_file(self, folder):
        """
        checks if the pickle file exist and loads it

        Parameters
        ----------
        folder: string
            folder name

        Returns
        -------
        data: numpy array
            data to append
        rawdata: numpy array
            data to append

        Examples
        --------
        >>> folder = '/2019-05-31-13-13-52/'
        >>> data = dataset.log_read_raw_data(folder = folder)
        >>> data.shape

        """
        from pickle import load
        with open(self.folder + 'experiment.pickle', 'rb') as file:
            data = load(file)
        return data

    def dump_to_pickle_file(self, obj = None):
        """
        pickles data and puts it on the drive. the default name is experiment.pickle (similar to original experiment.log)


        Parameters
        ----------
        data: numpy array
            data to append

        Returns
        -------

        Examples
        --------
        >>> dataset = Dataset()
        >>> dataset.dump_to_pickle_file(data = data)

        """
        from pickle import dump, HIGHEST_PROTOCOL
        if obj is not None:
            with open(self.folder + 'experiment.pickle', 'wb') as file:
                dump(obj = obj, file = file, protocol=HIGHEST_PROTOCOL)

    def get_lst(self,folder = '', type = 'cooling'):
        from numpy import genfromtxt
        import os
        dir_lst = os.listdir(folder + "buffer_files/")
        temp_lst = []
        for item in dir_lst:
            if '_'+type in item and '._' not in item:
                t_lst = item.split('_')
                t_lst.append(item)
                temp_lst.append(t_lst)
        lst2 = sorted(temp_lst, key=lambda x: int(x[1]))
        return lst2

    def combine_log_entries(self,raw_data):
        """
        combines all entries associated with one period in one entry.       converts raw log file data to collapsed 2D numpy array where every entry corresponds to one period. The time stamp on the period will be taken from the period event, which is defined elsewhere.

        Parameters
        ----------
        raw_data: numpy array
            raw data from original log file

        Returns
        -------
        data: numpy array
            compressed data

        Examples
        --------
        >>> folder = '/2019-05-31-13-13-52/'
        >>> raw_data = genfromtxt(folder + 'experiment.log', delimiter = ',', skip_header = 2)
        >>> data = dataset.combine_log_entries(raw_data)
        >>> data.shape

        """
        #find the max period
        max_period = int(max(raw_data[:,2]))
        data = []
        #step thpought every period index and every entry in the header
        for i in list(range(max_period+1)):
            self.i = i
            temp = []
            for j in range(len(self.log_header)):
                #col = 2 is period index
                value = raw_data[where(raw_data[:,2] == i),:][:,:,j][~isnan(raw_data[where(raw_data[:,2] == i),:][:,:,j])]
                if len(value) == 1:
                    temp.append(float(raw_data[where(raw_data[:,2] == i),:][:,:,j][~isnan(raw_data[where(raw_data[:,2] == i),:][:,:,j])]))
                elif len(value) == 0:
                    temp.append(nan)
                elif len(value) >1:
                    temp.append(float(raw_data[where(raw_data[:,2] == i),:][:,:,j][~isnan(raw_data[where(raw_data[:,2] == i),:][:,:,j])][0]))
            data.append(temp)
        return asarray(data)

    def get_log_vector(self,param = 'None'):
        """

        """
        from numpy import squeeze
        try:
            idx = self.log_header.index(param)
        except:
            print(f"param {param} doesn't exist")
            idx = None
        if not None:
            vector = self.period[:,idx]
        return vector

    def get_trace(self,period = 0, type = ''):
        """
        returns the trace numpy array from /buffer_files/

        Parameters
        ----------
        period: integer
            period index number
        type: string
            type of buffer file (pre,depre,pump,cooling, etc)

        Returns
        -------
        data: numpy array
            data to append

        Examples
        --------
        >>> data = dataset.get_trace(period = 2, type = 'pump')
        >>> data.shape
        """
        import os
        from numpy import genfromtxt, transpose
        filename = 'some_nonexisting_file.x'
        for item in self.trace_lists[type]:
            if item[1] == str(period):
                filename = item[3]
        filepath = self.folder + 'buffer_files/'+filename
        exists = os.path.isfile(filepath)
        if exists:
            data = transpose(genfromtxt(filepath,delimiter = ','))
        else:
            data = None
        return data


    def plot_trace(self,type = None, period = None, show = False):
        """
        returns matplotlib figure object of a trace of selected type and period. Returns None if tracefile doesn't exist.

        Parameters
        ----------
        type: string
            type of buffer file (pre,depre,pump,cooling, etc)
        period: integer
            period index number
        show: boolean
            optional plot show flag


        Returns
        -------
        object: matplotlib object
            matplotlib object

        Examples
        --------
        >>> data = dataset.get_trace(period = 2, type = 'pump')
        >>> data.shape
        """
        from matplotlib import pyplot as plt

        # fetch the data from the /buffers/ folder.
        if period is not None and type is not None:
            data = self.get_trace(type = type, period = period)
            if data is not None:
                pass
            else:
                pass
        else:
            return None

    def plot_log(self, type = None):
        """
        returns matplotlib figure object of a trace of selected type and period. Returns None if tracefile doesn't exist.

        Parameters
        ----------
        type: string
            type of buffer file (pre,depre,pump,cooling, etc)
        period: integer
            period index number
        show: boolean
            optional plot show flag


        Returns
        -------
        object: matplotlib object
            matplotlib object

        Examples
        --------
        >>> data = dataset.plot_log(type = 'all')
        >>> data.shape
        """
        pass



if __name__ == '__main__':
    from pdb import pm
    import os
    folder = get_test_folder()
    dataset = Dataset(folder)



"""
DATAQ 4108 driver level code
author: Valentyn Stadnytskyi
"""

from numpy import nan, mean, std, asarray, array, concatenate, \
     delete, round, vstack, hstack, zeros, transpose, split
from time import time, sleep

import sys
if sys.version_info[0] == 3:
    if sys.version_info[1] <= 7:
        from time import gmtime, strftime, time, sleep, clock
    else:
        from time import gmtime, strftime, time, sleep
        from time import perf_counter as clock
else:
    from time import gmtime, strftime, time, sleep, clock
import os.path
import struct
from pdb import pm
from time import gmtime, strftime
import logging
from struct import pack, unpack
from timeit import Timer
from logging import info,warn, debug, error
import traceback

mock_commands = {}
mock_commands['info 1 \r'] = 'info 1  4108\r'
mock_commands['info 2 \r'] = 'info 2  VS\r'
mock_commands['info 3 \r'] = 'info 3  4A\r'
mock_commands['info 4 \r'] = 'info 4  1234567890\r'
mock_commands['info 5 \r'] = 'info 5  Mary_Smith\r'
mock_commands['info 6 \r'] = 'info 6  00000000\r'
mock_commands['info 7 \r'] = 'info 7  00000000\r'
mock_commands['info 8 \r'] = 'info 8  1\r'
mock_commands['info 9 \r'] = 'info 9  60000000\r'
mock_commands['led 1 \r'] = 'led 1 \r'
mock_commands['led 2 \r'] = 'led 2 \r'
mock_commands['led 3 \r'] = 'led 3 \r'
mock_commands['led 4 \r'] = 'led 4 \r'
mock_commands['led 5 \r'] = 'led 5 \r'


class Driver(object):
    def __init__(self):
        self.available_ports = []
        self.dev = None
        self.sim_analog = [11000,0,8000,0,8000,11000,9000,0]
        self.sim_analog_std = [30,30,30,30,30,30,30,30]
        self.sim_digital = 127
        self.sim_last_read = time()
        self.sim_pressure_state = 1
        self.pump_freq = 1#0.99995
        self.sim_slow_leak_start_time = time()
        self.sim_slow_leak_tau = 60000

        self.sim_flag = False

        from icarus_nmr.tests.test_data.mock_driver import traces

        self.lst_pre = traces.get_lst_pre_trace()
        self.lst_depre = traces.get_lst_pre_trace()
        self.lst_pump = traces.get_lst_pump_trace()

    def init(self):
        self.discover()
        self.use_port()
        self.sim_pump_event_flag = False
        self.sim_pre_event_flag = False
        self.sim_depre_event_flag = False

        info("initialization of the driver is complete")

    def get_information(self):
        """
        emulates retrieval of information via usb. Creates two new atrbiutes to the class: self.dev_dict and self.epi_dict

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> mock.get_information()
        """
        if self.dev != None:
            self.dev_dict = {}
            self.epi_dict = {}
            self.dev_dict['DEV:address'] = 'DEV:address:simulator'
            self.dev_dict['DEV:bDeviceClass'] = 'DEV:bDeviceClass:simulator'
            self.dev_dict['DEV:bDescriptorType'] = 'DEV:bDescriptorType:simulator'
            self.dev_dict['DEV:bDeviceProtocol'] = 'DEV:bDeviceProtocol:simulator'
            self.dev_dict['DEV:bLength'] = 'DEV:bLength:simulator'
            self.dev_dict['DEV:bMaxPacketSize0'] = 'DEV:bMaxPacketSize0:simulator'
            self.dev_dict['DEV:bNumConfigurations'] = 0
            self.dev_dict['DEV:manufacturer'] = 0
            self.dev_dict['DEV:serial_number'] = 0
            self.dev_dict['DEV:speed'] = 0
            self.dev_dict['DEV:product'] = 0

            #endpoint IN description
            self.epi_dict['EPI:bmAttributes'] = 0
            self.epi_dict['EPI:wMaxPacketSize'] = 0
            self.epi_dict['EPI:bSynchAddress'] = 0
            self.epi_dict['EPI:bInterval'] = 0
            self.epi_dict['EPI:bEndpointAddress'] = 0
            self.epi_dict['EPI:bDescriptorType'] = 0
            self.epi_dict['EPI:bInterval'] = 0
            self.epi_dict['EPI:bInterval'] = 0

    def get_hardware_information(self):
        """
        emulates retrieval of information via usb. Creates two new atrbiutes to the class: self.dev_dict and self.epi_dict

        Parameters
        ----------

        Returns
        -------
        dic :: dictionary
            dictionary of hardware information

        Examples
        --------
        >>> dic = mock.get_hardware_information()
        >>> dic

        """

        dic = {}
        dic[b'Device name'] = self.inquire('info 1 \r').split('info 1 ')[1][1:-1]
        dic[b'Firmware version'] = self.inquire('info 2 \r').split('info 2 ')[1][1:-1]
        dic[b'Serial Number'] = self.inquire('info 6 \r').split('info 6 ')[1][1:-1]
        dic[b'Sample Rate Divisor'] = self.inquire('info 9 \r').split('info 9 ')[1][1:-1]

        return dic



    def use_port(self, port = ''):
        pass

    def discover(self):
        self.dev = 0


    """Set and Get persistent_property"""
    # functions for persistent properties if needed

    """Basic serial communication functions"""

    def read(self,N):

        while time() - self.sim_last_read <= N/4000.0:#512.0/4000.0:
            sleep(0.01)
        self.sim_last_read = time()
        return True

    def readall(self):
        """
        reads all information in the usb buffer

        Parameters
        ----------

        Returns
        -------
        buffer :: string
            buffer information as a string

        Examples
        --------
        >>> dic = mock.get_hardware_information()
        >>> dic
        """
        return self.read(0)

    def write(self,command):
        """
        writes into USB buffer

        Parameters
        ----------
        command :: string
            a string representation of a command

        Returns
        -------

        Examples
        --------
        >>> mock.write(command = 'info 1 \\r')
        """
        try:
            self.dev.write(self.epo,command)
        except:
            error(traceback.format_exc())

    def inquire(self,command):
        if command in list(mock_commands.keys()):
            res = mock_commands[command]
        else:
            res = 'Command not found\r\x00'
        return res

    def flush(self):
        pass

    def close_port(self):
        pass
    """DI-4108 Communcation functions"""

    def config_digital(self, number = 127, echo = False):
        #tested dec 17, 2017
        return True

    def parse_binary_number(self,value = 0, Nbits = 7):
        """
        parses integer number into bits and represents them as an array

        Parameters
        ----------
        value :: integer
            integer number

        Returns
        -------

        Examples
        --------
        >>> mock.write(command = 'info 1 \\r')
        """
        from numpy import arange,ndarray,nan
        binary = format(value, '#0'+str(Nbits+3)+'b')
        arr = arange(Nbits)
        for i in range(Nbits):
            arr[i] = binary[Nbits+2-i]
        return arr

    def set_digital(self, number = 127, echo = False):
        """
        sets digital value. The mock driver emulates pressure transitions due to change in digital state.

        Parameters
        ----------
        number :: integer
            integer number that will be set to DI-4108 digital outputs
        echo :: boolean, optional
            optional flag that turn on reply.

        Returns
        -------

        Examples
        --------
        >>> mock.set_digital(number = 127)
        """
        from time import time
        before = self.sim_digital
        if 0<=number<=127:
            after = number
            self.sim_digital = number
            bin_array = self.parse_binary_number(value = after) - self.parse_binary_number(value = before)
            if bin_array[1] == -1 and self.sim_pressure_state != 0:
                self.sim_depre_event_trace = self.get_depre_trace()
                self.sim_depre_event_idx = 0
                self.sim_depre_event_flag = True
                self.sim_pressure_state = 0
            if bin_array[2] == -1 and self.sim_pressure_state != 1:
                self.sim_pre_event_trace = self.get_pre_trace()
                self.sim_pre_event_idx = 0
                self.sim_pre_event_flag = True
                self.sim_pressure_state = 1
            if bin_array[2] == 1:
                self.sim_slow_leak_start_time = time()
            if bin_array[1] == -1:
                sim_pump_event_flag = True
            elif bin_array[1] == 1:
                self.sim_pump_event_flag = False
        else:
            pass

        if echo == True:
            flag = True
        else:
            flag = False
        return flag

    def set_analog(self, echo = False, filterN = 1):
        flag = True
        if echo:
            flag = True
        return flag

    def set_acquisition_rate(self, rate = 1000, baserate = 20000, dec = 0, echo = True):
        """
        sets acquisition rate (simulation rate)

        Parameters
        ----------
        rate :: integer
            integer number that will be set to DI-4108 digital outputs
        baserate :: boolean, optional
            optional flag that turn on reply.
        dec :: integer

        Returns
        -------

        Examples
        --------
        >>> mock.set_acquisition_rate(rate = 1000, baserate = 20000, dec = 4)
        """
        self.sim_rate = int(1.0*baserate/dec)
        return True

    def start_scan(self):
        self.sim_flag = True
        info( 'The configured measurement(s) has(have) started')

    def stop_scan(self):
        self.sim_flag = False
        info("measurement ended")

    def full_stop(self):
        self.stop_scan()
        self.close_port()

    """Advance function"""
    def config_channels(self, rate = 3000, conf_digital = 127, set_digital = 127, dec = 20, baserate = 20000):
        x = self.config_digital(conf_digital, echo = True)
        a = self.set_digital(set_digital, echo = True)
        b = self.set_analog(echo = True)
        c = self.set_acquisition_rate(rate = rate, dec = dec, baserate = 20000, echo = True) #3000 gives maximum rate
        return x*a*b*c

    def get_readings(self, points_to_read = 1, to_read_analog = 8, to_read_digital = 1):
        """
        the original get_readings function should just return data from the buffer on the DI-4108 DATAQ. Since this is module is a mock device, there are extra functionality build in into this specific function. The function handles random number generation around specified values with specified standard deviations. The fucntion also includes what happens if pressurization and depresurrization event flags are True.
        """
        from numpy import zeros, random
        from time import time

        # first generate empty packets of specified dimensionality
        to_read = (int(to_read_analog)+int(to_read_digital)*2)
        res = zeros((to_read,points_to_read),dtype = 'int16')
        t1 = time()
        k=0
        for j in range(points_to_read):
            ###pump event handler
            if self.sim_pump_event_flag:
                if self.sim_pump_event_idx<len(self.sim_pump_event_trace):
                    self.sim_analog[0] = self.sim_pump_event_trace[self.sim_pump_event_idx]
                    self.sim_analog_std[0] = 1
                    self.sim_pump_event_idx +=1
                else:
                    self.sim_pump_event_flag = False
                    self.sim_analog_std[0] = 10
            else:
                flag = random.uniform(0,1) > self.pump_freq
                if flag:
                    self.sim_pump_event_idx = 0
                    self.sim_pump_event_trace = self.get_pump_trace()
                    self.sim_pump_event_flag = True
                else:
                    self.sim_pump_event_flag = False
            if self.sim_pre_event_flag:
                if self.sim_pre_event_idx<len(self.sim_pre_event_trace[5,:]):
                    self.sim_analog[5] = self.sim_pre_event_trace[5,self.sim_pre_event_idx]
                    self.sim_analog[6] = self.sim_pre_event_trace[6,self.sim_pre_event_idx]
                    self.sim_analog_std[5] = 1
                    self.sim_analog_std[6] = 1
                    self.sim_pre_event_idx +=1
                else:
                    self.sim_pre_event_flag = False
                    self.sim_analog_std[5] = 10
                    self.sim_analog_std[6] = 10
            if self.sim_depre_event_flag:
                if self.sim_depre_event_idx<len(self.sim_depre_event_trace[5,:]):
                    self.sim_analog[5] = self.sim_depre_event_trace[5,self.sim_depre_event_idx]
                    self.sim_analog[6] = self.sim_depre_event_trace[6,self.sim_depre_event_idx]
                    self.sim_analog_std[5] = 1
                    self.sim_analog_std[6] = 1
                    self.sim_depre_event_idx +=1
                else:
                    self.sim_depre_event_flag = False
                    self.sim_analog_std[5] = 10
                    self.sim_analog_std[6] = 10
            for i in [0,1,2,3,4,7]:
                res[i,j] = random.normal(self.sim_analog[i],self.sim_analog_std[i],1)[0]
            for i in [5,6]:
                res[i,j] = random.normal(self.sim_analog[i]*self.slow_leak(),self.sim_analog_std[i],1)[0]
            res[9,j] = self.sim_digital

            # this section simulates continous data generation. It will sleep for ~16 ms every 64 points.
            k+=1
            if k >= 64:
                sleep(64/self.sim_rate)
                k = 0
        flag = True
        return res.T, flag #(analog_data,digital_data)

    def blink(self):
        pass

# Supporting
    def get_pre_trace(self):
        print('get pre trace')
        from numpy import arange, zeros
        import random

        N = int(random.uniform(0,len(self.lst_pre)-1))
        spl_list = self.lst_pre[N]
        x = arange(0,333,1)*0.25
        data = zeros((10,333))

        data[5,:] = spl_list[5](x)
        var_first = spl_list[5](x[0])
        data[5,:] = data[5,:]-var_first
        val_last = data[5,-1]
        data[5,:] = data[5,:]*12000/val_last

        data[6,:] = spl_list[6](x)
        var_first = spl_list[6](x[0])
        data[6,:] = data[6,:]-var_first
        val_last = data[6,-1]
        data[6,:] = data[6,:]*12000/val_last
        return data

    def get_depre_trace(self):
        print('get depre trace')
        from numpy import arange, zeros
        import random

        N = int(random.uniform(0,len(self.lst_depre)-1))
        spl_list = self.lst_depre[N]
        x = arange(0,333,1)*0.25
        data = zeros((10,333))

        data[5,:] = spl_list[5](x)
        var_first = spl_list[5](x[-1])
        data[5,:] = data[5,:]-var_first
        val_last = data[5,0]
        data[5,:] = data[5,:]*12000/val_last

        data[6,:] = spl_list[6](x)
        var_first = spl_list[6](x[-1])
        data[6,:] = data[6,:]-var_first
        val_last = data[6,0]
        data[6,:] = data[6,:]*12000/val_last
        return data

    def get_pump_trace(self):
        print('get pump trace')
        from numpy import arange, zeros
        import random

        N = int(random.uniform(0,len(self.lst_pump)-1))
        spl_list = self.lst_pump[N]
        x = arange(0,333,1)*0.25
        data = spl_list(x)
        return data

########################################################
##### Artificial events
########################################################

    def pinhole_leak(self):
        """
        creates a pin hole leak event. Values in channel 5,6 are set to 0.

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> mock.pinhole_leak()
        """
        self.sim_analog[5] = 0
        self.sim_analog[6] = 0
        self.sim_pressure_state = 0

    def slow_leak(self):
        """
        emulates slow leak

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> mock.pinhole_leak()
        """
        def exp(t,amp,tau):
            from numpy import exp
            return exp(-t/tau)
        #self.sim_slow_leak_flag = flag
        #self.sim_slow_leak_tau = 1
        #self.sim_slow_leak_start_time = 0
        return exp(t = time()-self.sim_slow_leak_start_time, amp = 1, tau = self.sim_slow_leak_tau)


if __name__ == "__main__": #for testing
    import logging
    from tempfile import gettempdir
    driver = Driver()
    self = driver


    logging.basicConfig(filename=gettempdir()+'/DI_4108_BULK_LL.log',
                        level=logging.DEBUG, format="%(asctime)s %(levelname)s: %(message)s")
    print('self.self_test()')
    print('self.test1()')
    print('self.test2()')
    print('self.echo_test1()')
    print('self.performance_test1()')
    print('self.performance_test2()')



import unittest
from numpy.testing import assert_array_equal
import sys
if sys.version_info[0] < 3:
    from time import clock
else:
    from time import perf_counter as clock

class MockTest(unittest.TestCase):
    def test_get_hardware_information(self):
        from ..driver_mock import Driver
        driver = Driver()
        driver.init()
        dic = {b'Device name': '4108',
                 b'Firmware version': 'VS',
                 b'Serial Number': '00000000',
                 b'Sample Rate Divisor': '60000000'}
        result = driver.get_hardware_information()
        self.assertEqual(dic,result)

    def test_configure(self):
        from ..driver_mock import Driver
        driver = Driver()
        driver.init()
        driver.set_acquisition_rate(dec = 5)
        self.assertEqual(driver.sim_rate,4000)
        driver.set_acquisition_rate(dec = 4)
        self.assertEqual(driver.sim_rate,5000)

    def test_inquire(self):
        from ..driver_mock import Driver
        driver = Driver()
        driver.init()
        reply = driver.inquire('info 1 \r')
        self.assertEqual('info 1  4108\r',reply)
        reply = driver.inquire('info 5 \r')
        self.assertEqual('info 5  Mary_Smith\r',reply)

    def test_start_stop_scan(self):
        from ..driver_mock import Driver
        driver = Driver()
        driver.init()
        self.assertEqual(driver.sim_flag,False)
        driver.start_scan()
        self.assertEqual(driver.sim_flag,True)
        driver.stop_scan()
        self.assertEqual(driver.sim_flag,False)

    def test_get_readings(self):
        from ..driver_mock import Driver
        driver = Driver()
        driver.init()
        driver.config_channels()
        driver.start_scan()
        data, flag = driver.get_readings(512)
        self.assertEqual(data.shape,(512,10))
        self.assertEqual(data.T[9].mean(),127.0)
        self.assertEqual(data.T[9].std(), 0.0)

    def test_set_digital(self):
        from ..driver_mock import Driver
        driver = Driver()
        driver.init()
        driver.config_channels()
        driver.start_scan()
        data, flag = driver.get_readings(512)
        self.assertEqual(data.T[9].mean(),127.0)

        driver.set_digital(126)
        data, flag = driver.get_readings(512)
        self.assertEqual(data.T[9].mean(),126.0)

        driver.set_digital(122)
        data, flag = driver.get_readings(512)
        self.assertEqual(data.T[9].mean(),122.0)


    def test_get_pre_trace(self):
        from ..driver_mock import Driver
        driver = Driver()
        driver.init()
        driver.config_channels()
        driver.start_scan()
        trace = driver.get_pre_trace()
        self.assertEqual(trace.shape, (10, 333))
        self.assertEqual(trace[5,0]<6000.0, True)
        self.assertEqual(trace[5,-1]<6000.0, False)


    # def test_1(self):
    #     from numpy import random
    #     from ..circular_buffer import CircularBuffer
    #     buffer = CircularBuffer(shape=(100, 2, 4))
    #     data = random.randint(1024, size=(5, 2, 4))
    #     buffer.packet_length = 5
    #     buffer.append(data)
    #     self.assertEqual(buffer.pointer, 4)
    #     self.assertEqual(buffer.g_pointer, 4)
    #     self.assertEqual(buffer.packet_pointer, 0)
    #     self.assertEqual(buffer.g_packet_pointer, 0)
    # def test_queue_end(self):
    #     """
    #     test if the default pointer in the buffer is -1.
    #     """
    #     from ..circular_buffer import CircularBuffer
    #     buffer = CircularBuffer(shape=(100, 2))
    #     self.assertEqual(buffer.pointer, -1)
    #
    # def test_queue_end_two(self):
    #     """
    #     test if the default pointer in the buffer is -1.
    #     """
    #     from ..circular_buffer import CircularBuffer
    #     buffer = CircularBuffer(shape=(100, 2))
    #     self.assertEqual(buffer.pointer, -1)
    #
    # def test_1(self):
    #     from numpy import random
    #     from ..circular_buffer import CircularBuffer
    #     buffer = CircularBuffer(shape=(100, 2, 4))
    #     data = random.randint(1024, size=(5, 2, 4))
    #     buffer.packet_length = 5
    #     buffer.append(data)
    #     self.assertEqual(buffer.pointer, 4)
    #     self.assertEqual(buffer.g_pointer, 4)
    #     self.assertEqual(buffer.packet_pointer, 0)
    #     self.assertEqual(buffer.g_packet_pointer, 0)
    #
    # def test_attributes(self):
    #     from ..circular_buffer import CircularBuffer
    #     from numpy import random
    #     buffer = CircularBuffer(shape=(100, 2), dtype='int16')
    #     data = random.randint(1024, size=(5, 2))
    #     buffer.append(data)
    #     self.assertEqual(buffer.shape, (100, 2))
    #     self.assertEqual(buffer.size, 100*2)
    #     self.assertEqual(buffer.dtype, 'int16')
    #
    # def test_full(self):
    #     from ..circular_buffer import CircularBuffer
    #     from numpy import random, sum
    #     buffer = CircularBuffer(shape=(100, 2, 3), dtype='float64')
    #     data = random.randint(1024, size=(50, 2, 3))
    #     buffer.append(data)
    #     assert buffer.pointer == 49
    #     assert buffer.g_pointer == 49
    #     assert buffer.shape == (100, 2, 3)
    #     assert buffer.size == buffer.buffer.shape[0]*buffer.buffer.shape[1]*buffer.buffer.shape[2]
    #     assert buffer.dtype == 'float64'
    #     assert sum(buffer.get_i_j(i=5, j=6)) == sum(buffer.buffer[5])
    #     # get data between pointers 5 and 10 and compare to get 5 points from pointer M
    #     assert sum(buffer.get_i_j(i=5, j=10)) == sum(buffer.get_N(N=5, M=9))
